#!/usr/bin/env python
"""
Unittests for the Resubmission spec factory
"""
from __future__ import print_function

import os
import unittest
from copy import deepcopy

import threading

from Utils.PythonVersion import PY3
from WMCore.DAOFactory import DAOFactory
from WMCore.Database.CMSCouch import CouchServer, Document
from WMCore.WMSpec.StdSpecs.ReReco import ReRecoWorkloadFactory
from WMCore.WMSpec.StdSpecs.Resubmission import ResubmissionWorkloadFactory
from WMCore.WMSpec.StdSpecs.StepChain import StepChainWorkloadFactory
from WMCore.WMSpec.StdSpecs.TaskChain import TaskChainWorkloadFactory
from WMCore.WMSpec.WMSpecErrors import WMSpecFactoryException
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.TestInitCouchApp import TestInitCouchApp


TASK2 = {'ConfigCacheID': 'FAKE',
         'GlobalTag': 'GT_DP_V1',
         'TaskName': 'RECODreHLT',
         'TimePerEvent': 123,
         'Memory': 2200,
         'Multicore': 22,
         'EventStreams': 2}

STEP2 = {'ConfigCacheID': 'FAKE',
         'GlobalTag': 'GT_DP_V1',
         'StepName': 'RECODreHLT',
         'Multicore': 22,
         'EventStreams': 2}


class ResubmissionTests(EmulatedUnitTestCase):
    """
    Unit tests for the Resubmission spec

    NOTE: test performance/resource values are actually overwritten by the
    original/parent workflow which we are creating an ACDC for
    """

    def setUp(self):
        """
        Initialize the database and couch.
        """
        super(ResubmissionTests, self).setUp()
        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection(destroyAllDatabase=True)
        self.testInit.setupCouch("resubmission_t", "ReqMgr")
        self.testInit.setupCouch("resubmission_t", "ConfigCache")
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)

        couchServer = CouchServer(os.environ["COUCHURL"])
        self.configDatabase = couchServer.connectDatabase("resubmission_t")
        self.testInit.generateWorkDir()

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.listTasksByWorkflow = self.daoFactory(classname="Workflow.LoadFromName")
        self.listFilesets = self.daoFactory(classname="Fileset.List")
        self.listSubsMapping = self.daoFactory(classname="Subscriptions.ListSubsAndFilesetsFromWorkflow")
        if PY3:
            self.assertItemsEqual = self.assertCountEqual
        return

    def tearDown(self):
        """
        Clear out the database.
        """
        self.testInit.clearDatabase()
        self.testInit.delWorkDir()
        super(ResubmissionTests, self).tearDown()

    def getDefaultACDCParams(self, origType):
        """
        Returns a default dictionary to be used in the Resubmission
        workflow creation.
        :param origType: spec type for the parent workflow (one being ACDC'ed)
        :return: a dictionary
        """
        defaultArgs = {"ACDCDatabase": "acdcserver",
                       "ACDCServer": self.testInit.couchUrl,
                       "CouchURL": "https://cmsweb-testbed.cern.ch/couchdb",
                       "RequestType": "Resubmission",
                       "OriginalRequestType": origType}
        return defaultArgs

    def testStandardTaskChainACDC(self):
        """
        Creates a standard ACDC for a TaskChain workflow with no parameter updates
        """
        testArguments = TaskChainWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("TaskChain")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_TaskChain_LumiMask_multiRun_HG2108_Val_210731_182010_1852/HLTD",
                         "OriginalRequestName": "amaltaro_TaskChain_LumiMask_multiRun_HG2108_Val_210731_182010_1852",
                         "RequestString": "ACDC_UnitTest_TaskChain_LumiMask_multiRun"})
        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TaskChain_ACDC",
                                                           testArguments, acdcArgs)

        # now we can validate the new workload object
        for topTask in testWorkload.taskIterator():
            for taskObj in topTask.taskIterator():
                perfParams = taskObj.jobSplittingParameters()['performance']
                print("Task: {}, type: {}, perf: {}".format(taskObj.name(), taskObj.taskType(), perfParams))
                if taskObj.taskType() in ('Production', 'Processing'):
                    for step in ('cmsRun1', 'stageOut1', 'logArch1'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams['memoryRequirement'], 2400.0)
                    self.assertEqual(perfParams['timePerEvent'], 0.8)
                else:
                    for step in taskObj.listAllStepNames():
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})

        # and now we validate the workflow high level description (JSON)
        # top level must have the default value for Resubmission creation (None)
        for param in ("TimePerEvent", "Memory", "Multicore", "EventStreams"):
            self.assertIsNone(testArguments[param])

    def testCustomTaskChainACDC(self):
        """
        Creates a custom ACDC for a TaskChain workflow overwriting some parameters
        """
        testArguments = TaskChainWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("TaskChain")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_TaskChain_LumiMask_multiRun_HG2108_Val_210731_182010_1852/HLTD",
                         "OriginalRequestName": "amaltaro_TaskChain_LumiMask_multiRun_HG2108_Val_210731_182010_1852",
                         "RequestString": "ACDC_UnitTest_TaskChain_LumiMask_multiRun",
                         "Memory": {"HLTD": 1950, "RECODreHLT": 2950.0},
                         "Multicore": {"HLTD": 11, "RECODreHLT": 12},
                         "EventStreams": {"HLTD": 21, "RECODreHLT": 22},
                         "TimePerEvent": {"HLTD": 1.5, "RECODreHLT": 2.5}})
        testArguments["Task2"] = deepcopy(TASK2)
        testArguments["Task1"]["TaskName"] = "HLTD"
        testArguments["TaskChain"] = 2
        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TaskChain_ACDC",
                                                           testArguments, acdcArgs)

        # now we can validate the new workload object
        for topTask in testWorkload.taskIterator():
            for taskObj in topTask.taskIterator():
                taskName = taskObj.name()
                perfParams = taskObj.jobSplittingParameters()['performance']
                print("Task: {}, type: {}, perf: {}".format(taskName, taskObj.taskType(), perfParams))
                if taskObj.taskType() in ('Production', 'Processing'):
                    stepHelper = taskObj.getStepHelper('cmsRun1')
                    self.assertEqual(stepHelper.getNumberOfCores(), acdcArgs['Multicore'][taskName])
                    self.assertEqual(stepHelper.getNumberOfStreams(), acdcArgs['EventStreams'][taskName])
                    for step in ('stageOut1', 'logArch1'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams['memoryRequirement'], acdcArgs['Memory'][taskName])
                    self.assertEqual(perfParams['timePerEvent'], acdcArgs['TimePerEvent'][taskName])
                elif taskObj.taskType() == 'LogCollect':
                    stepHelper = taskObj.getStepHelper('logCollect1')
                    self.assertEqual(stepHelper.getNumberOfCores(), 1)
                    self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})
                elif taskObj.taskType() == 'Cleanup':
                    for step in taskObj.listAllStepNames():
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})

        # and now we validate the workflow high level description (JSON)
        # top level must be equal to what is provided by the client
        for param in ("TimePerEvent", "Memory", "Multicore", "EventStreams"):
            self.assertItemsEqual(testArguments[param], acdcArgs[param])
            # and the task needs to be properly updated
            for taskNum in ("Task1", "Task2"):
                taskName = testArguments[taskNum]["TaskName"]
                self.assertEqual(testArguments[taskNum][param], acdcArgs[param][taskName])

    def testCustomSimpleTaskChainACDC(self):
        """
        Creates a custom ACDC for a TaskChain workflow overwriting some parameters
        """
        testArguments = TaskChainWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("TaskChain")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_TaskChain_LumiMask_multiRun_HG2108_Val_210731_182010_1852/HLTD",
                         "OriginalRequestName": "amaltaro_TaskChain_LumiMask_multiRun_HG2108_Val_210731_182010_1852",
                         "RequestString": "ACDC_UnitTest_TaskChain_LumiMask_multiRun",
                         "Memory": 9999,
                         "Multicore": 22,
                         "EventStreams": 11,
                         "TimePerEvent": 2.5})
        testArguments["Task2"] = deepcopy(TASK2)
        testArguments["Task1"]["TaskName"] = "HLTD"
        testArguments["TaskChain"] = 2
        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TaskChain_ACDC",
                                                           testArguments, acdcArgs)

        # now we can validate the new workload object
        for topTask in testWorkload.taskIterator():
            for taskObj in topTask.taskIterator():
                taskName = taskObj.name()
                perfParams = taskObj.jobSplittingParameters()['performance']
                print("Task: {}, type: {}, perf: {}".format(taskName, taskObj.taskType(), perfParams))
                if taskObj.taskType() in ('Production', 'Processing'):
                    stepHelper = taskObj.getStepHelper('cmsRun1')
                    self.assertEqual(stepHelper.getNumberOfCores(), acdcArgs['Multicore'])
                    self.assertEqual(stepHelper.getNumberOfStreams(), acdcArgs['EventStreams'])
                    for step in ('stageOut1', 'logArch1'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams['memoryRequirement'], acdcArgs['Memory'])
                    self.assertEqual(perfParams['timePerEvent'], acdcArgs['TimePerEvent'])
                elif taskObj.taskType() == 'LogCollect':
                    stepHelper = taskObj.getStepHelper('logCollect1')
                    self.assertEqual(stepHelper.getNumberOfCores(), 1)
                    self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})
                elif taskObj.taskType() == 'Cleanup':
                    for step in taskObj.listAllStepNames():
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})

        # and now we validate the workflow high level description (JSON)
        # top level must be equal to what is provided by the client
        for param in ("TimePerEvent", "Memory", "Multicore", "EventStreams"):
            self.assertEqual(testArguments[param], acdcArgs[param])
            # and the task needs to be properly updated
            for taskNum in ("Task1", "Task2"):
                self.assertEqual(testArguments[taskNum][param], acdcArgs[param])

    def testStandardStepChainACDC(self):
        """
        Creates a standard ACDC for a StepChain workflow
        """
        testArguments = StepChainWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("StepChain")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_SC_PY3_PURecyc_HG2109_Val_210903_160613_9070/RecoPU_2021PU",
                         "OriginalRequestName": "amaltaro_SC_PY3_PURecyc_HG2109_Val_210903_160613_9070",
                         "RequestString": "ACDC_UnitTest_SC_PY3_PURecyc"})
        testArguments["Step2"] = deepcopy(STEP2)
        testArguments["Step2"]["StepName"] = "Nano_2021PU"
        testArguments["StepChain"] = 2
        testArguments["Multicore"] = 8
        testArguments["EventStreams"] = 2
        testArguments["Memory"] = 10000
        testArguments["TimePerEvent"] = 120
        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("StepChain_ACDC",
                                                           testArguments, acdcArgs)
        # now we can validate the new workload object
        for topTask in testWorkload.taskIterator():
            for taskObj in topTask.taskIterator():
                perfParams = taskObj.jobSplittingParameters()['performance']
                print("Task: {}, type: {}, perf: {}".format(taskObj.name(), taskObj.taskType(), perfParams))
                if taskObj.taskType() in ('Production', 'Processing'):
                    for step in ('cmsRun1', 'cmsRun2'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 8)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 2)
                    for step in ('stageOut1', 'logArch1'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams['memoryRequirement'], 10000)
                    self.assertEqual(perfParams['timePerEvent'], 120)
                else:
                    for step in taskObj.listAllStepNames():
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})

        # and now we validate the workflow high level description (JSON)
        # first, parameters only supported at the top level dictionary
        self.assertEqual(testArguments["Memory"], 10000)
        self.assertEqual(testArguments["TimePerEvent"], 120)
        for stepNum in ("Step1", "Step2"):
            self.assertTrue("Memory" not in testArguments[stepNum])
            self.assertTrue("TimePerEvent" not in testArguments[stepNum])
        # and these are set to their default values (None), if not provided by the user
        self.assertIsNone(testArguments["Multicore"])
        self.assertIsNone(testArguments["EventStreams"])
        self.assertTrue("Multicore" not in testArguments["Step1"])
        self.assertTrue("EventStreams" not in testArguments["Step1"])
        self.assertEqual(testArguments["Step2"]["Multicore"], 22)
        self.assertEqual(testArguments["Step2"]["EventStreams"], 2)

    def testFailStepChainACDC(self):
        """
        Creates an ACDC for a StepChain workflow with disallowed parameters
        """
        testArguments = StepChainWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("StepChain")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_SC_PY3_PURecyc_HG2109_Val_210903_160613_9070/RecoPU_2021PU",
                         "OriginalRequestName": "amaltaro_SC_PY3_PURecyc_HG2109_Val_210903_160613_9070",
                         "RequestString": "ACDC_UnitTest_SC_PY3_PURecyc"})
        testArguments["Step2"] = deepcopy(STEP2)
        testArguments["Step1"]["StepName"] = "RecoPU_2021PU"
        testArguments["Step2"]["StepName"] = "Nano_2021PU"
        testArguments["StepChain"] = 2

        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()
        # Memory cannot be of dictionary type for ACDCs of StepChain
        testArguments["Memory"] = {"RecoPU_2021PU": 1950, "Nano_2021PU": 2950.0}
        with self.assertRaises(WMSpecFactoryException):
            factory.factoryWorkloadConstruction("StepChain_ACDC", testArguments, acdcArgs)

        testArguments["Memory"] = 2
        testArguments["TimePerEvent"] = {"RecoPU_2021PU": 1.5, "Nano_2021PU": 2.5}
        # Neither TimePerEvent can, for ACDCs of StepChain
        with self.assertRaises(WMSpecFactoryException):
            factory.factoryWorkloadConstruction("StepChain_ACDC", testArguments, acdcArgs)

        testArguments["TimePerEvent"] = 1.5
        # And now it should go through
        factory.factoryWorkloadConstruction("StepChain_ACDC", testArguments, acdcArgs)

    def testCustomStepChainACDC(self):
        """
        Creates a custom ACDC for a StepChain workflow overwriting some parameters
        """
        testArguments = StepChainWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("StepChain")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_SC_PY3_PURecyc_HG2109_Val_210903_160613_9070/RecoPU_2021PU",
                         "OriginalRequestName": "amaltaro_SC_PY3_PURecyc_HG2109_Val_210903_160613_9070",
                         "RequestString": "ACDC_UnitTest_SC_PY3_PURecyc",
                         "Memory": 1950,
                         "Multicore": {"RecoPU_2021PU": 11, "Nano_2021PU": 12},
                         "EventStreams": {"RecoPU_2021PU": 21, "Nano_2021PU": 22},
                         "TimePerEvent": 19.5})
        testArguments["Step2"] = deepcopy(STEP2)
        testArguments["Step1"]["StepName"] = "RecoPU_2021PU"
        testArguments["Step2"]["StepName"] = "Nano_2021PU"
        testArguments["StepChain"] = 2
        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("StepChain_ACDC",
                                                           testArguments, acdcArgs)
        # now we can validate the new workload object
        for topTask in testWorkload.taskIterator():
            for taskObj in topTask.taskIterator():
                perfParams = taskObj.jobSplittingParameters()['performance']
                print("Task: {}, type: {}, perf: {}".format(taskObj.name(), taskObj.taskType(), perfParams))
                if taskObj.taskType() in ('Production', 'Processing'):
                    stepHelper = taskObj.getStepHelper("cmsRun1")
                    self.assertEqual(stepHelper.getNumberOfCores(), acdcArgs["Multicore"]["RecoPU_2021PU"])
                    self.assertEqual(stepHelper.getNumberOfStreams(), acdcArgs["EventStreams"]["RecoPU_2021PU"])
                    stepHelper = taskObj.getStepHelper("cmsRun2")
                    self.assertEqual(stepHelper.getNumberOfCores(), acdcArgs["Multicore"]["Nano_2021PU"])
                    self.assertEqual(stepHelper.getNumberOfStreams(), acdcArgs["EventStreams"]["Nano_2021PU"])
                    for step in ('stageOut1', 'logArch1'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams['memoryRequirement'], acdcArgs["Memory"])
                    self.assertEqual(perfParams['timePerEvent'], acdcArgs["TimePerEvent"])
                elif taskObj.taskType() == "Harvesting":
                    # Harvesting task is allowed to get memory/TpE update
                    self.assertEqual(perfParams['memoryRequirement'], acdcArgs["Memory"])
                    self.assertEqual(perfParams['timePerEvent'], acdcArgs["TimePerEvent"])
                else:
                    for step in taskObj.listAllStepNames():
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})

        # and now we validate the workflow high level description (JSON)
        # first, parameters only supported at the top level dictionary
        self.assertEqual(testArguments["Memory"], acdcArgs["Memory"])
        self.assertEqual(testArguments["TimePerEvent"], acdcArgs["TimePerEvent"])
        for stepNum in ("Step1", "Step2"):
            self.assertTrue("Memory" not in testArguments[stepNum])
            self.assertTrue("TimePerEvent" not in testArguments[stepNum])
            stepName = testArguments[stepNum]["StepName"]
            self.assertEqual(testArguments[stepNum]["Multicore"], acdcArgs["Multicore"][stepName])
            self.assertEqual(testArguments[stepNum]["EventStreams"], acdcArgs["EventStreams"][stepName])
        # and these are set to their default values (None), if not provided by the user
        self.assertItemsEqual(testArguments["Multicore"], acdcArgs["Multicore"])
        self.assertItemsEqual(testArguments["EventStreams"], acdcArgs["EventStreams"])

    def testCustomSimpleStepChainACDC(self):
        """
        Creates a custom ACDC for a StepChain workflow overwriting some parameters
        """
        testArguments = StepChainWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("StepChain")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_SC_PY3_PURecyc_HG2109_Val_210903_160613_9070/RecoPU_2021PU",
                         "OriginalRequestName": "amaltaro_SC_PY3_PURecyc_HG2109_Val_210903_160613_9070",
                         "RequestString": "ACDC_UnitTest_SC_PY3_PURecyc",
                         "Memory": 1950,
                         "Multicore": 11,
                         "EventStreams": 21,
                         "TimePerEvent": 19.5})
        testArguments["Step2"] = deepcopy(STEP2)
        testArguments["Step1"]["StepName"] = "RecoPU_2021PU"
        testArguments["Step2"]["StepName"] = "Nano_2021PU"
        testArguments["StepChain"] = 2
        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("StepChain_ACDC",
                                                           testArguments, acdcArgs)
        # now we can validate the new workload object
        for topTask in testWorkload.taskIterator():
            for taskObj in topTask.taskIterator():
                perfParams = taskObj.jobSplittingParameters()['performance']
                print("Task: {}, type: {}, perf: {}".format(taskObj.name(), taskObj.taskType(), perfParams))
                if taskObj.taskType() in ('Production', 'Processing'):
                    for step in ('cmsRun1', 'cmsRun2'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), acdcArgs["Multicore"])
                        self.assertEqual(stepHelper.getNumberOfStreams(), acdcArgs["EventStreams"])
                    for step in ('stageOut1', 'logArch1'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams['memoryRequirement'], acdcArgs["Memory"])
                    self.assertEqual(perfParams['timePerEvent'], acdcArgs["TimePerEvent"])
                elif taskObj.taskType() == "Harvesting":
                    # Harvesting task is allowed to get memory/TpE update
                    self.assertEqual(perfParams['memoryRequirement'], acdcArgs["Memory"])
                    self.assertEqual(perfParams['timePerEvent'], acdcArgs["TimePerEvent"])
                else:
                    for step in taskObj.listAllStepNames():
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})

        # and now we validate the workflow high level description (JSON)
        # first, parameters only supported at the top level dictionary
        self.assertEqual(testArguments["Memory"], acdcArgs["Memory"])
        self.assertEqual(testArguments["TimePerEvent"], acdcArgs["TimePerEvent"])
        for stepNum in ("Step1", "Step2"):
            self.assertTrue("Memory" not in testArguments[stepNum])
            self.assertTrue("TimePerEvent" not in testArguments[stepNum])
            self.assertEqual(testArguments[stepNum]["Multicore"], acdcArgs["Multicore"])
            self.assertEqual(testArguments[stepNum]["EventStreams"], acdcArgs["EventStreams"])
        # and these are set to their default values (None), if not provided by the user
        self.assertEqual(testArguments["Multicore"], acdcArgs["Multicore"])
        self.assertEqual(testArguments["EventStreams"], acdcArgs["EventStreams"])

    def testStandardReRecoACDC(self):
        """
        Creates a standard ACDC for a ReReco workflow
        """
        testArguments = ReRecoWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("ReReco")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_ReReco_RunBlockWhite_HG2108_Val_210731_181900_4258/DataProcessing",
                         "OriginalRequestName": "amaltaro_ReReco_RunBlockWhite_HG2108_Val_210731_181900_4258",
                         "RequestString": "ACDC_UnitTest_ReReco_RunBlockWhite"})
        testArguments["Multicore"] = 8
        testArguments["EventStreams"] = 0
        testArguments["Memory"] = 12000
        testArguments["TimePerEvent"] = 73.85
        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("ReReco_ACDC",
                                                           testArguments, acdcArgs)
        # now we can validate the new workload object
        for topTask in testWorkload.taskIterator():
            for taskObj in topTask.taskIterator():
                perfParams = taskObj.jobSplittingParameters()['performance']
                print("Task: {}, type: {}, perf: {}".format(taskObj.name(), taskObj.taskType(), perfParams))
                if taskObj.taskType() in ('Production', 'Processing'):
                    stepHelper = taskObj.getStepHelper("cmsRun1")
                    self.assertEqual(stepHelper.getNumberOfCores(), testArguments["Multicore"])
                    self.assertEqual(stepHelper.getNumberOfStreams(), testArguments["EventStreams"])
                    for step in ('stageOut1', 'logArch1'):
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams['memoryRequirement'], testArguments["Memory"])
                    self.assertEqual(perfParams['timePerEvent'], testArguments["TimePerEvent"])
                else:
                    for step in taskObj.listAllStepNames():
                        stepHelper = taskObj.getStepHelper(step)
                        self.assertEqual(stepHelper.getNumberOfCores(), 1)
                        self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                    self.assertEqual(perfParams, {})

        # and now we validate the workflow high level description (JSON)
        # only top level parameters are allowed
        self.assertEqual(testArguments["Memory"], 12000)
        self.assertEqual(testArguments["TimePerEvent"], 73.85)
        self.assertEqual(testArguments["Multicore"], 8)
        self.assertEqual(testArguments["EventStreams"], 0)

    def testFailReRecoACDC(self):
        """
        Creates an ACDC for a ReReco workflow with disallowed parameters
        """
        testArguments = ReRecoWorkloadFactory.getTestArguments()
        # use a workflow existent in cmsweb-testbed, not good, but better than nothing
        acdcArgs = self.getDefaultACDCParams("ReReco")
        acdcArgs.update({"InitialTaskPath": "/amaltaro_ReReco_RunBlockWhite_HG2108_Val_210731_181900_4258/DataProcessing",
                         "OriginalRequestName": "amaltaro_ReReco_RunBlockWhite_HG2108_Val_210731_181900_4258",
                         "RequestString": "ACDC_UnitTest_ReReco_RunBlockWhite"})
        testArguments.update(acdcArgs)
        factory = ResubmissionWorkloadFactory()

        # Memory cannot be of dictionary type for ACDCs of StepChain
        testArguments["Memory"] = {"RecoPU_2021PU": 1950, "Nano_2021PU": 2950.0}
        with self.assertRaises(WMSpecFactoryException):
            factory.factoryWorkloadConstruction("ReReco_ACDC", testArguments, acdcArgs)

        testArguments["Memory"] = 2
        testArguments["TimePerEvent"] = {"RecoPU_2021PU": 1.5, "Nano_2021PU": 2.5}
        # Neither TimePerEvent can, for ACDCs of StepChain
        with self.assertRaises(WMSpecFactoryException):
            factory.factoryWorkloadConstruction("ReReco_ACDC", testArguments, acdcArgs)

        testArguments["TimePerEvent"] = 1.5
        # And now it should go through
        factory.factoryWorkloadConstruction("ReReco_ACDC", testArguments, acdcArgs)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
WorkQueue splitting for Resubmission workflows

In this case we can't be agnostic of the specific splitting algorithm
for the top level task. Each algorithm may require a different way of generating
ACDC blocks.

Current implementations:

- SingleChunk: Harvest, ParentlessMergeBySize, MinFileBased, EventAwareLumiBased, LumiBased
- FixedSizeChunks: FileBased, FixedDelay, MergeBySize, RunBased, SizeBased, SplitFileBased, TwoFileAndEventBased, TwoFileBased

ACDC unsupported:

- WMBSMergeBySize
- SiblingProcessingBased

"""
import json
from math import ceil
from WMCore.WorkQueue.Policy.Start.StartPolicyInterface import StartPolicyInterface
from WMCore.WorkQueue.WorkQueueExceptions import WorkQueueWMSpecError
from WMCore.WorkQueue.WorkQueueUtils import makeLocationsList
from WMCore.WorkQueue.DataStructs.ACDCBlock import ACDCBlock
from WMCore.ACDC.DataCollectionService import DataCollectionService


class ResubmitBlock(StartPolicyInterface):
    """Split elements into blocks"""

    def __init__(self, **args):
        StartPolicyInterface.__init__(self, **args)
        self.args.setdefault('SliceType', 'NumberOfFiles')
        self.args.setdefault('SliceSize', 1)
        self.args.setdefault('SplittingAlgo', 'LumiBased')
        self.lumiType = "NumberOfLumis"

        # Define how to handle the different splitting algorithms
        self.algoMapping = {'Harvest': self.singleChunk,
                            'ParentlessMergeBySize': self.singleChunk,
                            'MinFileBased': self.singleChunk,
                            'LumiBased': self.singleChunk,
                            'EventAwareLumiBased': self.singleChunk,
                            'EventBased': self.singleChunk}
        self.unsupportedAlgos = ['WMBSMergeBySize', 'SiblingProcessingBased']
        self.defaultAlgo = self.fixedSizeChunk
        self.sites = []

    def split(self):
        """Apply policy to spec"""
        # Prepare a site list in case we need it
        siteWhitelist = self.initialTask.siteWhitelist()
        siteBlacklist = self.initialTask.siteBlacklist()
        self.sites = makeLocationsList(siteWhitelist, siteBlacklist)

        for block in self.validBlocks(self.initialTask):
            parentList = {}
            parentFlag = False
            if self.initialTask.parentProcessingFlag():
                parentFlag = True
                parentList[block["Name"]] = block['Sites']

            self.newQueueElement(Inputs={block['Name']: block['Sites']},
                                 ParentFlag=parentFlag,
                                 ParentData=parentList,
                                 NumberOfLumis=block[self.lumiType],
                                 NumberOfFiles=block['NumberOfFiles'],
                                 NumberOfEvents=block['NumberOfEvents'],
                                 Jobs=ceil(float(block[self.args['SliceType']]) /
                                           float(self.args['SliceSize'])),
                                 ACDC=block['ACDC'],
                                 NoInputUpdate=self.initialTask.getTrustSitelists().get('trustlists'),
                                 NoPileupUpdate=self.initialTask.getTrustSitelists().get('trustPUlists')
                                )

    def validate(self):
        """Check args and spec work with block splitting"""
        StartPolicyInterface.validateCommon(self)

    def validBlocks(self, task):
        """Return blocks that pass the input data restriction according
           to the splitting algorithm"""
        validBlocks = []

        acdcInfo = task.getInputACDC()
        if not acdcInfo:
            raise WorkQueueWMSpecError(self.wmspec, 'No acdc section for %s' % task.getPathName())
        acdc = DataCollectionService(acdcInfo["server"], acdcInfo["database"])

        self.logger.info("Policy self.data variable content: %s", self.data)
        if self.data:
            acdcBlockSplit = ACDCBlock.splitBlockName(next(iter(self.data)))
        else:
            # if self.data is not passed, assume the data is input dataset from the spec
            acdcBlockSplit = False

        self.logger.info("Using ACDC blockSplit:\n%s", json.dumps(acdcBlockSplit, indent=2))
        if acdcBlockSplit:
            block = acdc.getChunkInfo(acdcInfo['collection'],
                                      acdcBlockSplit['TaskName'],
                                      acdcBlockSplit['Offset'],
                                      acdcBlockSplit['NumOfFiles'])
            dbsBlock = {}
            dbsBlock['Name'] = next(iter(self.data))

            dbsBlock['NumberOfFiles'] = block['files']
            dbsBlock['NumberOfEvents'] = block['events']
            dbsBlock['NumberOfLumis'] = block['lumis']
            dbsBlock['ACDC'] = acdcInfo
            if task.getTrustSitelists().get('trustlists'):
                dbsBlock["Sites"] = self.sites
            else:
                dbsBlock["Sites"] = self.cric.PNNstoPSNs(block["locations"])
            validBlocks.append(dbsBlock)
        else:
            if self.args['SplittingAlgo'] in self.unsupportedAlgos:
                raise WorkQueueWMSpecError(self.wmspec, 'ACDC is not supported for %s' % self.args['SplittingAlgo'])
            splittingFunc = self.defaultAlgo
            if self.args['SplittingAlgo'] in self.algoMapping:
                splittingFunc = self.algoMapping[self.args['SplittingAlgo']]
            validBlocks = splittingFunc(acdc, acdcInfo, task)
        self.logger.info("ACDC with the following validBlocks summary:\n%s", json.dumps(validBlocks, indent=2))

        return validBlocks

    def fixedSizeChunk(self, acdc, acdcInfo, task):
        """Return a set of blocks with a fixed number of ACDC records"""
        fixedSizeBlocks = []
        chunkSize = 250
        acdcBlocks = acdc.chunkFileset(acdcInfo['collection'],
                                       acdcInfo['fileset'],
                                       chunkSize)
        for block in acdcBlocks:
            dbsBlock = {}
            dbsBlock['Name'] = ACDCBlock.name(self.wmspec.name(),
                                              acdcInfo["fileset"],
                                              block['offset'], block['files'])
            dbsBlock['NumberOfFiles'] = block['files']
            dbsBlock['NumberOfEvents'] = block['events']
            dbsBlock['NumberOfLumis'] = block['lumis']
            if task.getTrustSitelists().get('trustlists'):
                dbsBlock["Sites"] = self.sites
            else:
                dbsBlock["Sites"] = self.cric.PNNstoPSNs(block["locations"])
            dbsBlock['ACDC'] = acdcInfo
            if dbsBlock['NumberOfFiles']:
                fixedSizeBlocks.append(dbsBlock)
        return fixedSizeBlocks

    def singleChunk(self, acdc, acdcInfo, task):
        """Return a single block (inside a list) with all associated ACDC records"""
        result = []
        acdcBlock = acdc.singleChunkFileset(acdcInfo['collection'],
                                            acdcInfo['fileset'])
        dbsBlock = {}
        dbsBlock['Name'] = ACDCBlock.name(self.wmspec.name(),
                                          acdcInfo["fileset"],
                                          acdcBlock['offset'], acdcBlock['files'])
        dbsBlock['NumberOfFiles'] = acdcBlock['files']
        dbsBlock['NumberOfEvents'] = acdcBlock['events']
        dbsBlock['NumberOfLumis'] = acdcBlock['lumis']
        if task.getTrustSitelists().get('trustlists'):
            dbsBlock["Sites"] = self.sites
        else:
            dbsBlock["Sites"] = self.cric.PNNstoPSNs(acdcBlock["locations"])
        dbsBlock['ACDC'] = acdcInfo
        if dbsBlock['NumberOfFiles']:
            result.append(dbsBlock)

        return result



import unittest
from WMCore.Services.WMArchive.CMSSWMetrics import CMSSWMetrics, CMSSW_METRICS


class CMSSWMetrics_t(unittest.TestCase):
    """
    Unit test class for CMSSW metrics module
    """

    def testCMSSWMetrics(self):
        """
        unit test for CMSSWMetrics function
        """
        tdict = CMSSWMetrics()
        for key, tdict in tdict.items():
            vdict = CMSSW_METRICS[key]
            if key != 'XrdSiteStatistics':
                for tkey, val in tdict.items():
                    value = vdict[tkey]
                    self.assertTrue(val == type(value))

    def testXrdMetrics(self):
        """
        unit test for XrdMetrics function
        """
        tdict = CMSSWMetrics()['XrdSiteStatistics']
        for key, xarr in tdict.items():
            varr = CMSSW_METRICS['XrdSiteStatistics'][key]
            # loop over each dict in specific metrics array
            for tdict in varr:
                for tkey, val in tdict.items():
                    etype = xarr[0][tkey]
                    # special case when expected type is float but value is 0 which interpreted as int
                    if val == 0 and etype == float:
                        continue
                    self.assertTrue(type(val) == etype)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Unittests for MemoryCache object
"""

import unittest
from time import sleep

from Utils.MemoryCache import MemoryCache, MemoryCacheException
#from Utils.PythonVersion import PY3


class MemoryCacheTest(unittest.TestCase):
    """
    unittest for MemoryCache functions
    """

    def setUp(self):
        self.assertItemsEqual = self.assertCountEqual

    def testBasics(self):
        cache = MemoryCache(1, [])
        self.assertItemsEqual(cache.getCache(), [])
        cache.setCache(["item1", "item2"])
        self.assertItemsEqual(cache.getCache(), ["item1", "item2"])
        # wait for cache to expiry, wait for 2 secs
        sleep(2)
        self.assertRaises(MemoryCacheException, cache.getCache)
        cache.setCache(["item4"])
        # and the cache is alive again
        self.assertItemsEqual(cache.getCache(), ["item4"])

    def testCacheSet(self):
        cache = MemoryCache(2, set())
        self.assertItemsEqual(cache.getCache(), set())
        cache.setCache(set(["item1", "item2"]))
        self.assertItemsEqual(cache.getCache(), ["item1", "item2"])
        cache.addItemToCache("item3")
        self.assertItemsEqual(cache.getCache(), ["item1", "item2", "item3"])
        cache.addItemToCache(["item4"])
        self.assertItemsEqual(cache.getCache(), ["item1", "item2", "item3", "item4"])
        cache.addItemToCache(set(["item5"]))
        self.assertItemsEqual(cache.getCache(), ["item1", "item2", "item3", "item4", "item5"])
        self.assertTrue("item2" in cache)
        self.assertFalse("item222" in cache)

    def testCacheList(self):
        cache = MemoryCache(2, [])
        self.assertItemsEqual(cache.getCache(), [])
        cache.setCache(["item1", "item2"])
        self.assertItemsEqual(cache.getCache(), ["item1", "item2"])
        cache.addItemToCache("item3")
        self.assertItemsEqual(cache.getCache(), ["item1", "item2", "item3"])
        cache.addItemToCache(["item4"])
        self.assertItemsEqual(cache.getCache(), ["item1", "item2", "item3", "item4"])
        cache.addItemToCache(set(["item5"]))
        self.assertItemsEqual(cache.getCache(), ["item1", "item2", "item3", "item4", "item5"])
        self.assertTrue("item2" in cache)
        self.assertFalse("item222" in cache)

    def testCacheDict(self):
        cache = MemoryCache(2, {})
        self.assertItemsEqual(cache.getCache(), {})
        cache.setCache({"item1": 11, "item2": 22})
        self.assertItemsEqual(cache.getCache(), {"item1": 11, "item2": 22})
        cache.addItemToCache({"item3": 33})
        self.assertItemsEqual(cache.getCache(), {"item1": 11, "item2": 22, "item3": 33})
        self.assertTrue("item2" in cache)
        self.assertFalse("item222" in cache)
        # test exceptions
        self.assertRaises(TypeError, cache.addItemToCache, "item4")
        self.assertRaises(TypeError, cache.addItemToCache, ["item4"])

    def testSetDiffTypes(self):
        cache = MemoryCache(2, set())
        self.assertItemsEqual(cache.getCache(), set())
        cache.setCache({"item1", "item2"})
        self.assertRaises(TypeError, cache.setCache, ["item3"])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
Scram_t.py

Created by Dave Evans on 2012-01-27.
Copyright (c) 2012 evansde77. All rights reserved.
"""

import unittest
import os
import tempfile

from WMQuality.TestInit import TestInit
from Utils.TemporaryEnvironment import tmpEnv
from WMCore.WMRuntime.Tools.Scram import (Scram, OS_TO_ARCH, ARCH_TO_OS, getSingleScramArch,
                                          isCMSSWSupported, isEnforceGUIDInFileNameSupported)


class Scram_t(unittest.TestCase):
    def setUp(self):
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testDir = self.testInit.generateWorkDir()
        self.oldCwd = os.getcwd()

    def tearDown(self):
        self.testInit.delWorkDir()

    def testA(self):
        """
        instantiate a Scram instance in test mode.
        """
        try:
            Scram(
                initialise="/bin/date",
                architecture="slc5_amd64_gcc454",
                version="CMSSW_X_Y_Z",
                test=True
            )
        except Exception as ex:
            msg = "Failed to instantiate Scram in test mode:\n %s " % str(ex)
            self.fail(msg)

    def testB(self):
        """
        instantiante a Scram instance in non-test mode
        limited what we can test here since we dont have scram etc in unittest env
        """
        try:
            Scram(
                initialise="/bin/date",
                architecture="slc5_amd64_gcc454",
                version="CMSSW_X_Y_Z"
            )
        except Exception as ex:
            msg = "Failed to instantiate Scram:\n %s " % str(ex)
            self.fail(msg)

    def testC(self):
        """
        test all method calls in test mode

        """
        s = Scram(
            initialise="/bin/date",
            architecture="slc5_amd64_gcc454",
            version="CMSSW_X_Y_Z",
            directory=self.testDir,
            test=True

        )

        try:
            status = s.project()
        except Exception as ex:
            msg = "Error running Scram.project:\n %s" % str(ex)
            self.fail(msg)

        self.assertEqual(status, 0)
        self.assertTrue(os.path.exists(s.projectArea))
        self.assertTrue("project" in s.lastExecuted)
        self.assertTrue("CMSSW_X_Y_Z" in s.lastExecuted)

        try:
            status = s.runtime()
        except Exception as ex:
            msg = "Error running Scram.runtime:\n %s" % str(ex)
            self.fail(msg)

        self.assertEqual(status, 0)
        self.assertTrue("ru -sh" in s.lastExecuted)
        self.assertTrue("TEST_MODE" in s.runtimeEnv)

        comm = "echo \"Hello World\""
        try:
            status = s(comm)
        except Exception as ex:
            msg = "Failed to call Scram object:\n %s" % str(ex)

        self.assertEqual(status, 0)
        self.assertEqual(s.lastExecuted, comm)

    def testArchMap(self):
        self.assertCountEqual(OS_TO_ARCH['rhel6'], ['slc5', 'slc6'])
        self.assertCountEqual(OS_TO_ARCH['rhel7'], ['slc7'])
        self.assertCountEqual(OS_TO_ARCH['rhel8'], ['el8', 'cc8', 'cs8', 'alma8'])
        self.assertCountEqual(OS_TO_ARCH['rhel9'], ['el9', 'cs9'])
        self.assertCountEqual(ARCH_TO_OS['slc6'], ['rhel6'])
        self.assertEqual(len(ARCH_TO_OS), 9)
        self.assertCountEqual(ARCH_TO_OS['slc7'], ['rhel7'])
        self.assertCountEqual(ARCH_TO_OS['slc7'], ['rhel7'])

    def testScramArchParsing(self):
        """
        Test the various modes of parsing for the scram arch
        """
        try:
            os.chdir(self.testDir)
            with tempfile.NamedTemporaryFile() as tf:
                tf.write(b'GLIDEIN_REQUIRED_OS = "rhel6" \n')
                tf.write(b'Memory = 2048\n')
                tf.flush()
                with tmpEnv(_CONDOR_MACHINE_AD=tf.name):
                    self.assertEqual(getSingleScramArch('slc6_blah_blah'), 'slc6_blah_blah')
                    self.assertEqual(getSingleScramArch('slc5_blah_blah'), 'slc5_blah_blah')
                    self.assertEqual(getSingleScramArch(['slc6_blah_blah', 'slc7_blah_blah']),
                                      'slc6_blah_blah')
                    self.assertEqual(getSingleScramArch(['slc6_blah_blah', 'slc5_blah_blah']),
                                      'slc6_blah_blah')
                    self.assertEqual(getSingleScramArch(['slc7_blah_blah', 'slc8_blah_blah']), None)
            with tempfile.NamedTemporaryFile() as tf:
                tf.write(b'GLIDEIN_REQUIRED_OS = "rhel7" \n')
                tf.write(b'Memory = 2048\n')
                tf.flush()
                with tmpEnv(_CONDOR_MACHINE_AD=tf.name):
                    self.assertEqual(getSingleScramArch('slc6_blah_blah'), 'slc6_blah_blah')
                    self.assertEqual(getSingleScramArch('slc7_blah_blah'), 'slc7_blah_blah')
                    self.assertEqual(getSingleScramArch(['slc6_blah_blah', 'slc7_blah_blah']),
                                      'slc7_blah_blah')
                    self.assertEqual(getSingleScramArch(['slc6_blah_blah', 'slc5_blah_blah']), None)
                    self.assertEqual(getSingleScramArch(['slc7_blah_blah', 'slc8_blah_blah']),
                                      'slc7_blah_blah')
        except Exception:
            raise
        finally:
            os.chdir(self.oldCwd)
        return

    def testCMSSWSupported(self):
        """
        Test the functionality of isCMSSWSupported function
        """
        self.assertFalse(isCMSSWSupported('CMSSW_1_2_3', ''))
        self.assertFalse(isCMSSWSupported(None, 'a'))
        self.assertFalse(isCMSSWSupported('CMSSW_1_2_3', 'CMSSW_2_2_3'))
        self.assertFalse(isCMSSWSupported('CMSSW_1_2_3', 'CMSSW_1_3_3'))
        self.assertFalse(isCMSSWSupported('CMSSW_1_2_3', 'CMSSW_1_2_4'))
        self.assertFalse(isCMSSWSupported('CMSSW_1_2_3_pre1', 'CMSSW_1_2_3'))
        self.assertFalse(isCMSSWSupported('CMSSW_1_2_3', 'CMSSW_1_2_3_pre1'))
        self.assertFalse(isCMSSWSupported('CMSSW_1_2_3_pre1', 'CMSSW_1_2_3_pre2'))
        self.assertFalse(isCMSSWSupported('CMSSW_1_2_3_pre2', 'CMSSW_1_2_3_pre1'))
        self.assertFalse(isCMSSWSupported('CMSSW_7_1_25_patch2', 'CMSSW_7_6_0'))
        self.assertFalse(isCMSSWSupported('CMSSW_7_3_2', 'CMSSW_10_4_0'))

        self.assertTrue(isCMSSWSupported('CMSSW_1_2_3_pre1', 'CMSSW_1_2_3_pre1'))
        self.assertTrue(isCMSSWSupported('CMSSW_1_2_3', 'CMSSW_1_2_3'))
        self.assertTrue(isCMSSWSupported('CMSSW_2_2_3', 'CMSSW_1_2_3'))
        self.assertTrue(isCMSSWSupported('CMSSW_1_3_3', 'CMSSW_1_2_3'))
        self.assertTrue(isCMSSWSupported('CMSSW_1_2_4', 'CMSSW_1_2_3'))

    def testisEnforceGUIDInFileNameSupported(self):
        """
        Test functionality of the `isEnforceGUIDInFileNameSupported` function
        """
        ### invalid input
        self.assertFalse(isEnforceGUIDInFileNameSupported(None))
        self.assertFalse(isEnforceGUIDInFileNameSupported(''))

        ### forever supported
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_11_0_0'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_11_0_2'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_11_1_0_pre1'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_11_1_0_patch1'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_11_1_1'))

        ### specific releases
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_10_2_20_UL'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_9_4_16_UL'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_8_0_34_UL'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_7_1_45_patch3'))

        ### minor supported releases
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_10_6_8'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_10_6_9'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_10_6_8_patch1'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_10_6_9_patch1'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_10_2_20'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_9_4_16'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_9_3_17'))
        self.assertTrue(isEnforceGUIDInFileNameSupported('CMSSW_8_0_34'))

        ### releases not supported
        self.assertFalse(isEnforceGUIDInFileNameSupported('CMSSW_10_6_7'))
        self.assertFalse(isEnforceGUIDInFileNameSupported('CMSSW_10_7_0'))
        self.assertFalse(isEnforceGUIDInFileNameSupported('CMSSW_10_2_19'))
        self.assertFalse(isEnforceGUIDInFileNameSupported('CMSSW_10_3_10'))
        self.assertFalse(isEnforceGUIDInFileNameSupported('CMSSW_5_3_10'))


if __name__ == '__main__':
    unittest.main()



"""
Pre-requisites:
 1. a valid proxy in your X509_USER_PROXY variable
 2. wmagent env: /data/srv/wmagent/current/apps/wmagent/etc/profile.d/init.sh

Script used during cmsweb-testbed validation (or a new agent validation).
It will:
 1. clone the WMCore repository
 2. fetch all templates available in WMCore/test/data/ReqMgr/requests/*
 3. create a new request for each of them (based on the parameters given
 in the command line)
 4. assign workflows during creation time (also based on the parameters
 provided in command line)
"""


from builtins import range, str as newstr, bytes as newbytes

import sys
import os
import pwd
import shlex
import json
import argparse
import logging
from subprocess import call
from time import sleep, time


def makeSiteWhitelist(jsonName, siteList):
    """
    Provided a template json file name and the site white list from
    the command line options; return the correct site white list based
    on some silly rules
    """
    if 'LHE_PFN' in jsonName:
        siteList = ["T1_US_FNAL"]
        print("Overwritting SiteWhitelist to: %s" % siteList)
    elif 'LHE' in jsonName or 'DQMHarvest' in jsonName:
        siteList = ["T2_CH_CERN"]
        print("Overwritting SiteWhitelist to: %s" % siteList)
    return siteList


def cloneRepo(logger):
    """
    Clone the WMCore repository, if needed
    """
    if os.path.isdir('WMCore'):
        logger.info("WMCore directory found. I'm not going to clone it again.")
        logger.info("You have 5 secs to abort this operation or live with that forever...\n")
        sleep(5)
    else:
        # Cloning WMCore repo
        command = ["git", "clone", "https://github.com/dmwm/WMCore.git"]
        try:
            retcode = call(command)
            if retcode == 0:
                logger.info("WMCore repository successfully cloned!")
            else:
                logger.error("Failed to clone WMCore. Error: %s ", retcode)
                sys.exit(1)
        except OSError as e:
            logger.error("Exception while cloning WMCore. Details: %s", str(e))
            sys.exit(2)


def parseArgs():
    """
    Well, parse the arguments passed in the command line :)
    """
    parser = argparse.ArgumentParser(description="Inject and/or assign requests in reqmgr/2")

    parser.add_argument('-c', '--campaign', required=True,
                        help='Campaign injected during the workflow creation')
    parser.add_argument('-r', '--reqStr', required=True,
                        help='Request string appended to the request name')
    parser.add_argument('-m', '--mode', default='DMWM',
                    help='Base directory for the templates. Options are: DMWM, Integration or Static')
    parser.add_argument('-f', '--filename',
                        help='Specifies a single file for injection (extesion must be included)')
    parser.add_argument('-u', '--url', default='https://cmsweb-testbed.cern.ch',
                        help='Url to inject the requests against')
    parser.add_argument('-t', '--team', default='testbed-vocms0230',
                        help='Team name of the agent for assignment')
    parser.add_argument('-s', '--site', default=["T1_US_FNAL", "T2_CH_CERN"],
                        help='Site white list for assignment')
    parser.add_argument('-a', '--acqEra', default='DMWM_TEST',
                        help='AcquisitionEra for assignment')
    parser.add_argument('-p', '--procStr', default='TEST_Alan_DS3',
                        help='ProcessingString for assignment')
    parser.add_argument('-v', '--procVer', default=11,
                        help='ProcessingVersion for assignment')
    parser.add_argument('-i', '--injectOnly', action='store_true', default=False,
                        help='Only injects requests but do not assign them')
    parser.add_argument('--dryRun', action='store_true', default=False,
                        help='Simulation mode only')
    parser.add_argument('--debug', action='store_true', default=False,
                        help='Set logging to debug mode.')
    args = parser.parse_args()

    # sites argument could be "T1_US_FNAL,T2_CH_CERN" ...
    if isinstance(args.site, (newstr, newbytes)):
        args.site = args.site.split(',')

    return args


def handleAssignment(args, fname, jsonData):
    """
    Tweak the assignment parameters, if needed.
    We only overwrite assignment parameters that are already present
    in the json template
    """
    assignRequest = {}
    assignRequest.setdefault('Team', args.team)
    assignRequest.setdefault('Dashboard', "integration")
    assignRequest.setdefault('SiteWhitelist', makeSiteWhitelist(fname, args.site))
    # merge template name and current procStr, to avoid dups
    tmpProcStr = fname.replace('.json', '_') + args.procStr
    if 'AcquisitionEra' in jsonData['assignRequest']:
        assignRequest.setdefault('AcquisitionEra', args.acqEra)
    if 'ProcessingString' in jsonData['assignRequest']:
        assignRequest.setdefault('ProcessingString', tmpProcStr)
    if 'ProcessingVersion' in jsonData['assignRequest']:
        assignRequest.setdefault('ProcessingVersion', args.procVer)

    # dict args for TaskChain and StepChain
    if jsonData['createRequest']['RequestType'] in ["TaskChain", "StepChain"]:
        requestType = jsonData['createRequest']['RequestType']
        joker = requestType.split("Chain")[0]
        assignDict = jsonData['assignRequest']

        # reuse values provided in the request schema at creation level
        if 'AcquisitionEra' in assignDict and isinstance(assignDict['AcquisitionEra'], dict):
            assignRequest['AcquisitionEra'] = assignDict['AcquisitionEra']
            createDict = jsonData['createRequest']
            for i in range(1, createDict[requestType] + 1):
                innerDictName = "%s%d" % (joker, i)  # Task1, Task2, Step1, Step2 ...
                innerDict = createDict[innerDictName]
                # if there is no Task/Step level value, get it from the main dict
                assignRequest['AcquisitionEra'][innerDict["%sName" % joker]] = innerDict.get('AcquisitionEra', createDict['AcquisitionEra'])

        # always overwrite it as provided in the command line and task/step name
        if 'ProcessingString' in assignDict and isinstance(assignDict['ProcessingString'], dict):
            assignRequest['ProcessingString'] = assignDict['ProcessingString']
            for task, _ in assignRequest['ProcessingString'].items():
                assignRequest['ProcessingString'][task] = task + '_' + tmpProcStr

        # also reuse values as provided in the request schema at creation level
        if 'ProcessingVersion' in assignDict and isinstance(assignDict['ProcessingVersion'], dict):
            assignRequest['ProcessingVersion'] = assignDict['ProcessingVersion']
            createDict = jsonData['createRequest']
            for i in range(1, createDict[requestType] + 1):
                innerDictName = "%s%d" % (joker, i)  # Task1, Task2, Step1, Step2 ...
                innerDict = createDict[innerDictName]
                # if there is no Task/Step level value, get it from the main dict
                assignRequest['ProcessingVersion'][innerDict["%sName" % joker]] = innerDict.get('ProcessingVersion', createDict['ProcessingVersion'])

    jsonData['assignRequest'].update(assignRequest)
    return

def loggerSetup(logLevel=logging.INFO):
    """
    Return a logger which writes everything to stdout.
    """
    logger = logging.getLogger(__name__)
    outHandler = logging.StreamHandler(sys.stdout)
    outHandler.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s:%(module)s: %(message)s"))
    outHandler.setLevel(logLevel)
    logger.addHandler(outHandler)
    logger.setLevel(logLevel)
    return logger

def main():
    """
    Util to create and assign requests based on the templates available
    in WMCore repository.

    NOTE: it will inject and assign ALL templates under DMWM or Integration folder
    """
    args = parseArgs()
    if args.debug:
        logger = loggerSetup(logging.DEBUG)
    else:
        logger = loggerSetup()

    cloneRepo(logger)

    os.chdir("WMCore/test/data/ReqMgr")
    wmcorePath = "requests/" + args.mode + "/"
    if args.filename:
        # then only specified template will be injected
        if os.path.isfile(wmcorePath + args.filename):
            templates = [args.filename]
        else:
            logger.info("File %s not found.", wmcorePath + args.filename)
            templates = []
    else:
        templates = os.listdir(wmcorePath)

    # Filter out templates not allowed to be injected
    disallowedList = ['ReReco_badBlocks.json', 'TaskChain_InclParents.json',
                      'StepChain_InclParents.json', 'SC_Straight.json',
                      'StoreResults.json', 'Resub_MonteCarlo_eff.json', 'Resub_TaskChain_Multicore.json']
    logger.info("Skipping injection for these templates: %s\n", disallowedList)
    templates = [item for item in templates if item not in disallowedList]
    if not templates:
        logging.info("There are no templates to be injected.")
        sys.exit(3)

    startT = time()
    reqMgrCommand = "reqmgr2.py"

    # Temporary place to write the tweaked templates
    tmpFile = '/tmp/%s.json' % pwd.getpwuid(os.getuid()).pw_name
    wfCounter = 0

    # figure out which python version it is
    pythonCmd = "python3" if sys.version_info[0] == 3 else "python"

    for fname in templates:
        logger.info("Processing template: %s", fname)
        strComand = "%s %s -u %s -f %s -i " % (pythonCmd, reqMgrCommand, args.url, tmpFile)

        # read the original json template
        with open(wmcorePath + fname) as fo:
            jsonData = json.load(fo)

        # tweak the create dict
        createRequest = {}
        createRequest['RequestString'] = fname.split('.json')[0] + '_' + args.reqStr
        if 'Campaign' in jsonData['createRequest']:
            createRequest['Campaign'] = args.campaign
        jsonData['createRequest'].update(createRequest)

        # apply assignment overrides
        if not args.injectOnly:
            strComand += "-g "
            handleAssignment(args, fname, jsonData)

        # Dump the modified json in a temp file and use just it
        with open(tmpFile, "w") as outfile:
            json.dump(jsonData, outfile)

        if args.dryRun:
            logger.info("dry-run command: %s", strComand)
            continue

        # Inject and/or assign the request, for real
        injectComand = shlex.split(strComand)
        retcode = call(injectComand)
        if retcode == 0:
            logger.info("%s request successfully created!", fname)
            wfCounter += 1
        else:
            logger.info("%s request FAILED injection!", fname)

    totalT = time() - startT

    logger.info("\nInjected %d workflows out of %d templates in %.2f secs. Good job!",
                 wfCounter, len(templates), totalT)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Unittests for IteratorTools functions
"""

from __future__ import division, print_function

import unittest

from WMCore.ReqMgr.DataStructs.RequestError import InvalidSpecParameterValue
from WMCore.ReqMgr.DataStructs.RequestStatus import ACTIVE_STATUS, get_modifiable_properties
from WMCore.ReqMgr.Utils.Validation import (validateOutputDatasets,
                                            validate_request_priority,
                                            _validate_request_allowed_args)
from WMCore.WMSpec.StdSpecs.StdBase import StdBase


class ValidationTests(unittest.TestCase):
    """
    unittest for ReqMgr Utils Validation functions
    """

    def testValidateOutputDatasets(self):
        """
        Test the validateOutputDatasets function
        """
        dbsUrl = 'https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader/'

        outputDsets = ['/PD1/AcqEra1-ProcStr1-v1/GEN']
        self.assertIsNone(validateOutputDatasets(outputDsets, dbsUrl))
        outputDsets.append('/PD1/AcqEra1-ProcStr1-v1/GEN-SIM')
        self.assertIsNone(validateOutputDatasets(outputDsets, dbsUrl))
        outputDsets.append('/PD1/AcqEra1-ProcStr1-v1/GEN-SIM-RAW')
        self.assertIsNone(validateOutputDatasets(outputDsets, dbsUrl))

        outputDsets.append('/PD1/AcqEra1-ProcStr1-v1/GEN')
        with self.assertRaises(InvalidSpecParameterValue):
            validateOutputDatasets(outputDsets, dbsUrl)

        outputDsets.remove('/PD1/AcqEra1-ProcStr1-v1/GEN')
        outputDsets.append('/PD1//AOD')
        with self.assertRaises(InvalidSpecParameterValue):
            validateOutputDatasets(outputDsets, dbsUrl)

        outputDsets.remove('/PD1//AOD')
        outputDsets.append('/PD1/None/AOD')
        with self.assertRaises(InvalidSpecParameterValue):
            validateOutputDatasets(outputDsets, dbsUrl)

        outputDsets.remove('/PD1/None/AOD')
        outputDsets.append('/PD1/AcqEra1-ProcStr1-v1/ALAN')
        with self.assertRaises(InvalidSpecParameterValue):
            validateOutputDatasets(outputDsets, dbsUrl)

    def testRequestPriorityValidation(self):
        """
        Test the `validate_request_priority` function, which validates the
        RequestPriority parameter
        :return: nothing, raises an exception if there are problems
        """
        # test valid cases, integer in the range of [0, 1e6]
        for goodPrio in [0, 100, int(1e6 - 1)]:
            reqArgs = {'RequestPriority': goodPrio}
            print(reqArgs)
            validate_request_priority(reqArgs)

        # test invalid ranges
        for badPrio in [-10, 1e6, 1e7]:
            reqArgs = {'RequestPriority': badPrio}
            with self.assertRaises(InvalidSpecParameterValue):
                validate_request_priority(reqArgs)

        # test invalid data types
        for badPrio in ["1234", 1234.35, 1e6, [123]]:
            reqArgs = {'RequestPriority': badPrio}
            with self.assertRaises(InvalidSpecParameterValue):
                validate_request_priority(reqArgs)


    def testValidateRequestAllowedArgs(self):
        """
        Tests the `_validate_request_allowed_args` functions, which validates two pairs
        of request arguments and returns the difference between them and on top of that
        applies a filter of allowed parameters changes per status
        :return: nothing, raises an exception if there are problems
        """
        defReqArgs = StdBase.getWorkloadAssignArgs()
        newReqArgs = {key: None for key in defReqArgs.keys()}

        for status in ACTIVE_STATUS:
            # NOTE: We need to add the RequestStatus artificially and assign it
            #       to the currently tested active status
            defReqArgs["RequestStatus"] = status
            expectedReqArgs = {key: None for key in get_modifiable_properties(status)}
            reqArgsDiff = _validate_request_allowed_args(defReqArgs, newReqArgs)
            print(f"reqArgsDiff: {reqArgsDiff}")
            print(f"expectedReqArgs: {expectedReqArgs}")
            self.assertDictEqual(reqArgsDiff, expectedReqArgs)
            print("===============================")


if __name__ == '__main__':
    unittest.main()



"""
Use this for only unit test
"""
import os
import tempfile

from WMCore.Configuration import Configuration, saveConfigurationFile


def deleteConfig(configFile):
    if os.path.exists(configFile):
        os.remove(configFile)
    else:
        pass


def setupWMAgentConfig():
    _, configFile = tempfile.mkstemp(".py", "TESTAGENTConfig", )
    os.environ["WMAGENT_CONFIG"] = configFile
    _wmAgentConfig(configFile)
    return configFile


def _wmAgentConfig(configFile):
    config = Configuration()
    config.section_("General")
    config.General.logdb_name = "unittest_logdb"
    config.General.central_logdb_url = "http://localhost/central_logdb"
    config.General.ReqMgr2ServiceURL = "http://localhost/reqmgr2"

    config.section_("JobStateMachine")
    # Waring setting couchDB to None will cause the ERROR:
    # but that should be ignored, if you want to test couchDB
    # set the real couchDB information here
    config.JobStateMachine.couchurl = os.getenv("COUCHURL")
    config.JobStateMachine.couchDBName = os.getenv("COUCHDB")
    config.JobStateMachine.jobSummaryDBName = "wmagent_summary_test"
    config.JobStateMachine.summaryStatsDBName = "stat_summary_test"

    config.section_("Agent")
    # User specific parameter
    config.Agent.hostName = "cmssrv52.fnal.gov"
    # User specific parameter
    config.Agent.contact = "sfoulkes@fnal.gov"
    # User specific parameter
    config.Agent.teamName = "DMWM"
    # User specific parameter
    config.Agent.agentName = "WMAgent"
    config.Agent.useMsgService = False
    config.Agent.useTrigger = False
    config.Agent.isDocker = False

    # BossAir setup
    config.section_("BossAir")
    config.BossAir.pluginNames = ['TestPlugin', 'SimpleCondorPlugin']
    config.BossAir.pluginDir = 'WMCore.BossAir.Plugins'

    saveConfigurationFile(config, configFile)



import cherrypy
import re
import logging

class YUIServer:
    def __init__(self, config):
        self.yuidir = config.yuidir
        self.pattern = re.compile(r"^[-a-z_/]+\.(js|css|png)$")

    @cherrypy.expose
    def yui(self, *args):
        path = "/".join(args)
        if not self.pattern.match(path):
            logging.error("Bad YUI location "+path)
            raise cherrypy.HTTPError(403, "Bad YUI location")
        return cherrypy.lib.static.serve_file(self.yuidir + '/' + path)



#!/usr/bin/env python
"""
_InsertWorkflow_

MySQL implementation of DBSBuffer.InsertWorkflow
"""

from WMCore.Database.DBFormatter import DBFormatter

class InsertWorkflow(DBFormatter):
    """
    _InsertWorkflow_

    Insert a workflow using the name and task
    """

    sql = """INSERT IGNORE INTO dbsbuffer_workflow (name, task,
                                                    block_close_max_wait_time,
                                                    block_close_max_files,
                                                    block_close_max_events,
                                                    block_close_max_size)
                VALUES (:name, :task,
                        :blockMaxCloseTime,
                        :blockMaxFiles,
                        :blockMaxEvents,
                        :blockMaxSize)"""

    existsSQL = "SELECT id FROM dbsbuffer_workflow WHERE name = :name AND task = :task"


    def execute(self, requestName, taskPath,
                blockMaxCloseTime, blockMaxFiles,
                blockMaxEvents, blockMaxSize,
                conn = None, transaction = False):
        """
        _execute_

        Insert a simple workflow into the dbsbuffer_workflow table
        """
        binds = {'name': requestName, 'task': taskPath,
                 'blockMaxCloseTime' : blockMaxCloseTime,
                 'blockMaxFiles' : blockMaxFiles,
                 'blockMaxEvents' : blockMaxEvents,
                 'blockMaxSize' : blockMaxSize}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        binds = {'name': requestName, 'task': taskPath}

        result = self.dbi.processData(self.existsSQL, binds, conn = conn,
                                      transaction = transaction)
        workflowId = self.formatDict(result)[0]['id']
        return workflowId



#!/usr/bin/env python
"""
_NewWorkflow_

Oracle implementation of NewWorkflow
"""

from WMCore.WMBS.MySQL.Users.New import New as NewUserMySQL

class New(NewUserMySQL):
    sql = """INSERT INTO wmbs_users (id, cert_dn, name_hn, owner, grp, group_name, role_name)
             SELECT wmbs_users_SEQ.nextval, :dn, :hn, :owner, :grp, :gr, :role FROM dual
             WHERE NOT EXISTS (SELECT id FROM wmbs_users WHERE cert_dn = :dn
                                                         AND group_name = :gr
                                                         AND role_name = :role)"""


    def execute(self, dn, hn = None, owner = None, group = None,
                group_name = '', role_name = '',
                conn = None, transaction = False):

        binds = {"dn": dn, "hn": hn, "owner": owner, "grp": group,
                 "gr": group_name, "role": role_name}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        result = self.dbi.processData( self.sql_get_id, {'dn': dn,
                                                         "gr": group_name,
                                                         "role": role_name},
                                       conn = conn, transaction = transaction)
        id = self.format(result)
        return int(id[0][0])



#!/usr/bin/env python
"""
_StepFactory_

Unittests for StepFactory module

"""

import unittest

import WMCore.WMSpec.Steps.StepFactory as StepFactory


class StepFactoryTest(unittest.TestCase):
    """
    TestCase for StepFactory module. Load CMSSW implementation from
    each factory to test the mechanics

    """

    def testA(self):
        """template"""
        try:
            cmssw = StepFactory.getStepTemplate("CMSSW")
        except Exception as ex:
            msg = "Error loading Step Template of Type CMSSW\n"
            msg += str(ex)
            self.fail(msg)

    def testB(self):
        """bad template"""
        self.assertRaises(ImportError,
                          StepFactory.getStepTemplate, "UtterGuff")

    def testC(self):
        """builder"""
        try:
            cmssw = StepFactory.getStepBuilder("CMSSW")
        except Exception as ex:
            msg = "Error loading Step Builder of Type CMSSW\n"
            msg += str(ex)
            self.fail(msg)

    def testD(self):
        """bad buildere"""
        self.assertRaises(ImportError,
                          StepFactory.getStepBuilder, "UtterGuff")

    def testExecutor(self):
        """executor"""
        try:
            cmssw = StepFactory.getStepExecutor("CMSSW")
        except Exception as ex:
            msg = "Error loading Step Executor of Type CMSSW\n"
            msg += str(ex)
            self.fail(msg)

    def testBadExecutor(self):
        """bad executor"""
        self.assertRaises(ImportError,
                          StepFactory.getStepExecutor, "UtterGuff")

    def testEmulator(self):
        """emulator"""
        try:
            cmssw = StepFactory.getStepEmulator("CMSSW")
        except Exception as ex:
            msg = "Error loading Step Emulator of Type CMSSW\n"
            msg += str(ex)
            self.fail(msg)

    def testBadEmulator(self):
        """bad emu"""
        self.assertRaises(ImportError,
                          StepFactory.getStepEmulator, "UtterGuff")


if __name__ == '__main__':
    unittest.main()



"""
_EventAwareLumiBased_t_

Lumi based splitting test with awareness of events per lumi, using the DataStructs classes.
It must pass the same tests as the LumiBased algorithm, plus
specific ones for this algorithm.
See WMCore/WMBS/JobSplitting/ for the WMBS (SQL database) version.

Created on Sep 25, 2012

@author: dballest
"""
from builtins import range
from future.utils import viewvalues

import unittest

from WMCore.DataStructs.File import File
from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.Run import Run
from WMCore.DataStructs.Subscription import Subscription
from WMCore.DataStructs.Workflow import Workflow
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID


class EventAwareLumiBasedTest(unittest.TestCase):
    """
    _EventAwareLumiBasedTest_

    Test event based job splitting.
    """

    def setUp(self):
        """
        _setUp_

        Create two subscriptions: One that contains a single file and one that
        contains multiple files.
        """

        self.testWorkflow = Workflow()
        self.performanceParams = {'timePerEvent': 12,
                                  'memoryRequirement': 2300,
                                  'sizePerEvent': 400}

        return

    def tearDown(self):
        """
        _tearDown_

        Nothing to do...
        """
        pass

    def createSubscription(self, nFiles, lumisPerFile, twoSites=False, nEventsPerFile=100):
        """
        _createSubscription_

        Create a subscription for testing
        """

        baseName = makeUUID()

        testFileset = Fileset(name=baseName)
        for i in range(nFiles):
            newFile = self.createFile('%s_%i' % (baseName, i), nEventsPerFile,
                                      i, lumisPerFile, 'blenheim')
            testFileset.addFile(newFile)
        if twoSites:
            for i in range(nFiles):
                newFile = self.createFile('%s_%i_2' % (baseName, i), nEventsPerFile,
                                          i, lumisPerFile, 'malpaquet')
                testFileset.addFile(newFile)

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")

        return testSubscription

    def createFile(self, lfn, events, run, lumis, location):
        """
        _createFile_

        Create a file for testing
        """
        newFile = File(lfn=lfn, size=1000,
                       events=events)
        lumiList = []
        for lumi in range(lumis):
            lumiList.append((run * lumis) + lumi)
        newFile.addRun(Run(run, *lumiList))
        newFile.setLocation(location)
        return newFile

    def testA_FileSplitting(self):
        """
        _FileSplitting_

        Test that things work if we split files between jobs
        """
        splitter = SplitterFactory()

        oneSetSubscription = self.createSubscription(nFiles=10, lumisPerFile=1)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=oneSetSubscription)

        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               events_per_job=100,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs), 10)
        for job in jobGroups[0].jobs:
            self.assertTrue(len(job['input_files']), 1)

        twoLumiFiles = self.createSubscription(nFiles=5, lumisPerFile=2)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=twoLumiFiles)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               events_per_job=50,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        self.assertEqual(len(jobGroups[0].jobs), 10)
        for job in jobGroups[0].jobs:
            self.assertEqual(len(job['input_files']), 1)

        wholeLumiFiles = self.createSubscription(nFiles=5, lumisPerFile=3)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=wholeLumiFiles)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               events_per_job=67,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        # 10 because we split on run boundaries
        self.assertEqual(len(jobGroups[0].jobs), 10)
        jobList = jobGroups[0].jobs
        for job in jobList:
            # Have should have one file, half two
            self.assertTrue(len(job['input_files']) in [1, 2])

        mask0 = jobList[0]['mask'].getRunAndLumis()
        self.assertEqual(mask0, {0: [[0, 1]]})
        mask1 = jobList[1]['mask'].getRunAndLumis()
        self.assertEqual(mask1, {0: [[2, 2]]})
        mask2 = jobList[2]['mask'].getRunAndLumis()
        self.assertEqual(mask2, {1: [[3, 4]]})
        mask3 = jobList[3]['mask'].getRunAndLumis()
        self.assertEqual(mask3, {1: [[5, 5]]})

        self.assertEqual(jobList[0]['mask'].getRunAndLumis(), {0: [[0, 1]]})

        # Do it with multiple sites
        twoSiteSubscription = self.createSubscription(nFiles=5, lumisPerFile=2, twoSites=True)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=twoSiteSubscription)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               events_per_job=50,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 2)
        self.assertEqual(len(jobGroups[0].jobs), 10)
        for job in jobGroups[0].jobs:
            self.assertEqual(len(job['input_files']), 1)

    def testB_NoRunNoFileSplitting(self):
        """
        _NoRunNoFileSplitting_

        Test the splitting algorithm in the odder fringe
        cases that might be required.
        """
        splitter = SplitterFactory()
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=60,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 9)

        # The first job should have three lumis from one run
        # The second three lumis from two different runs
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 2]]})
        job1runLumi = jobs[1]['mask'].getRunAndLumis()
        self.assertEqual(job1runLumi[0][0][0] + 1, job1runLumi[0][0][1])  # Run 0, startLumi+1 == endLumi
        self.assertEqual(job1runLumi[1][0][0], job1runLumi[1][0][1])  # Run 1, startLumi == endLumi

        # Assert that this works differently with file splitting on and run splitting on
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=60,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 10)

        # In this case it should slice things up so that each job only has one run
        # in it.
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 2]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {0: [[3, 4]]})

        # Test total_events limit. (The algorithm cuts off after the lumi that
        # brings the total average event count over -or equal to- total_events.)
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=60,
                               total_events=10,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 1)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 0]]})

        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=60,
                               total_events=179,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 3)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 2]]})
        job1runLumi = jobs[1]['mask'].getRunAndLumis()
        self.assertEqual(job1runLumi[0][0][0] + 1, job1runLumi[0][0][1])  # Run 0, startLumi+1 == endLumi
        self.assertEqual(job1runLumi[1][0][0], job1runLumi[1][0][1])  # Run 1, startLumi == endLumi
        self.assertEqual(len(jobs[2]['mask'].getRunAndLumis()), 1)  # All lumis from one run again

        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=60,
                               total_events=180,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 3)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 2]]})
        job1runLumi = jobs[1]['mask'].getRunAndLumis()
        self.assertEqual(job1runLumi[0][0][0] + 1, job1runLumi[0][0][1])  # Run 0, startLumi+1 == endLumi
        self.assertEqual(job1runLumi[1][0][0], job1runLumi[1][0][1])  # Run 1, startLumi == endLumi
        self.assertEqual(len(jobs[2]['mask'].getRunAndLumis()), 1)  # All lumis from one run again

        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=60,
                               total_events=181,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 4)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 2]]})
        job1runLumi = jobs[1]['mask'].getRunAndLumis()
        self.assertEqual(job1runLumi[0][0][0] + 1, job1runLumi[0][0][1])  # Run 0, startLumi+1 == endLumi
        self.assertEqual(job1runLumi[1][0][0], job1runLumi[1][0][1])  # Run 1, startLumi == endLumi
        self.assertEqual(len(jobs[2]['mask'].getRunAndLumis()), 1)  # All lumis from one run again
        self.assertEqual(len(jobs[3]['mask'].getRunAndLumis()), 1)  # All lumis from one run again

        return

    def testC_FileSplitNoHardLimit(self):
        """
        _testC_FileSplitNoHardLimit_

        Simplest use case, there is only a self limit of events per job which
        the algorithm must adapt to on a file by file basis. At most
        one file per job so we don't have to pass information between files.
        """
        splitter = SplitterFactory()

        # Create 5 files with 7 lumi per file and 100 events per lumi on average.
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=7, twoSites=False,
                                                   nEventsPerFile=700)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        # First test, the optimal settings are 360 events per job
        # As we have files with 100 events per lumi, this will configure the splitting to
        # 3.6 lumis per job, which rounds to 3, the algorithm always approximates to the lower integer.
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=360,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 15, "There should be 15 jobs")

        # Now set the average to 200 events per job
        # This results in the algorithm reducing the lumis per job to 2
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=200,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 20, "There should be 20 jobs")

        # Check extremes, process a zero event files with lumis. It must be processed in one job
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=100, twoSites=False,
                                                   nEventsPerFile=0)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               events_per_job=5000,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 5, "There should be 5 jobs")

        # Process files with 10k events per lumi, fallback to one lumi per job. We can't do better
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=5, twoSites=False,
                                                   nEventsPerFile=50000)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=5000,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 25, "There should be 5 jobs")

        # Test total_events limit. (The algorithm cuts off after the lumi that
        # brings the total average event count over -or equal to- total_events.)
        testSubscription = self.createSubscription(nFiles=5, lumisPerFile=3, twoSites=False,
                                                   nEventsPerFile=300)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=250,
                               total_events=750,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 5, "There should be 5 jobs")
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 1]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {0: [[2, 2]]})
        self.assertEqual(jobs[2]['mask'].getRunAndLumis(), {1: [[3, 4]]})
        self.assertEqual(jobs[3]['mask'].getRunAndLumis(), {1: [[5, 5]]})
        self.assertEqual(len(jobs[4]['mask'].getRunAndLumis()), 1)  # All lumis from one run

        return

    def testD_NoFileSplitNoHardLimit(self):
        """
        _testD_NoFileSplitNoHardLimit_

        In this case we don't split on file boundaries, check different combination of files
        make sure we make the most of the splitting, e.g. include many zero event files in
        a single job.
        """
        splitter = SplitterFactory()

        # Create 100 files with 7 lumi per file and 0 events per lumi on average.
        testSubscription = self.createSubscription(nFiles=100, lumisPerFile=7, twoSites=False,
                                                   nEventsPerFile=0)
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        # First test, the optimal settings are 360 events per job
        # As we have files with 0 events per lumi, this will configure the splitting to
        # a single job containing all files
        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=360,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 1, "There should be 1 job")
        self.assertEqual(len(jobs[0]['input_files']), 100, "All 100 files must be in the job")

        # Create 7 files, each one with different lumi/event distributions
        testFileset = Fileset(name="FilesetA")
        testFileA = self.createFile("/this/is/file1", 250, 0, 5, "blenheim")  # job1, job2
        testFileB = self.createFile("/this/is/file2", 600, 1, 1, "blenheim")  # job3
        testFileC = self.createFile("/this/is/file3", 1200, 2, 2, "blenheim")  # job4, job5
        testFileD = self.createFile("/this/is/file4", 100, 3, 1, "blenheim")  # job6
        testFileE = self.createFile("/this/is/file5", 30, 4, 1, "blenheim")  # job6
        testFileF = self.createFile("/this/is/file6", 10, 5, 1, "blenheim")  # job6
        testFileG = self.createFile("/this/is/file7", 151, 6, 3, "blenheim")  # job7
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)
        testFileset.addFile(testFileD)
        testFileset.addFile(testFileE)
        testFileset.addFile(testFileF)
        testFileset.addFile(testFileG)

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")

        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        # Optimal settings are: jobs with 150 events per job
        # This means, the first file must be splitted in 3 lumis per job which would leave room
        # for another lumi in the second job, but the second file has a lumi too big for that
        # The 3rd job only contains the second file, the fourth and fifth job split the third file
        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=150,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 7, "Seven jobs must be in the jobgroup")
        self.assertEqual(jobs[0]["mask"].getRunAndLumis(), {0: [[0, 2]]}, "Wrong mask for the first job")
        self.assertEqual(jobs[1]["mask"].getRunAndLumis(), {0: [[3, 4]]}, "Wrong mask for the second job")
        self.assertEqual(jobs[2]["mask"].getRunAndLumis(), {1: [[1, 1]]}, "Wrong mask for the third job")
        self.assertEqual(jobs[3]["mask"].getRunAndLumis(), {2: [[4, 4]]}, "Wrong mask for the fourth job")
        self.assertEqual(jobs[4]["mask"].getRunAndLumis(), {2: [[5, 5]]}, "Wrong mask for the fifth job")
        self.assertEqual(jobs[5]["mask"].getRunAndLumis(),
                         {3: [[3, 3]], 4: [[4, 4]], 5: [[5, 5]]}, "Wrong mask for the sixth job")
        self.assertEqual(jobs[6]["mask"].getRunAndLumis(), {6: [[18, 20]]}, "Wrong mask for the seventh job")
        # Test interactions of this algorithm with splitOnRun = True
        # Make 2 files, one with 3 runs and a second one with the last run of the first
        fileA = File(lfn="/this/is/file1", size=1000,
                     events=2400)
        lumiListA = []
        lumiListB = []
        lumiListC = []
        for lumi in range(8):
            lumiListA.append(1 + lumi)
            lumiListB.append(1 + lumi)
            lumiListC.append(1 + lumi)
        fileA.addRun(Run(1, *lumiListA))
        fileA.addRun(Run(2, *lumiListA))
        fileA.addRun(Run(3, *lumiListA))
        fileA.setLocation("malpaquet")

        fileB = self.createFile('/this/is/file2', 200, 3, 5, "malpaquet")

        testFileset = Fileset(name='FilesetB')
        testFileset.addFile(fileA)
        testFileset.addFile(fileB)
        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")

        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        # The settings for this splitting are 700 events per job
        jobGroups = jobFactory(splitOnRun=True,
                               halt_job_on_file_boundaries=False,
                               events_per_job=700,
                               performance=self.performanceParams)
        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 6, "Six jobs must be in the jobgroup")

    def testE_HardLimitSpltting(self):
        """
        _testE_HardLimitSplitting_

        Test that we can specify a event limit, the
        algorithm shall take single lumi files with more events than the limit
        and mark them for failure
        """
        splitter = SplitterFactory()

        # Create 3 files, the one in the middle is a "bad" file
        testFileset = Fileset(name="FilesetA")
        testFileA = self.createFile("/this/is/file1", 1000, 0, 5, "blenheim")
        testFileB = self.createFile("/this/is/file2", 1000, 1, 1, "blenheim")
        testFileC = self.createFile("/this/is/file3", 1000, 2, 2, "blenheim")
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        # Settings are to split on job boundaries, to fail sing lumis with more than 800 events
        # and to put 550 events per job
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=550,
                               job_time_limit=9600,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 6, "Six jobs must be in the jobgroup")
        failedJobs = [job for job in jobs if job.get('failedOnCreation', False)]
        self.assertEqual(len(failedJobs), 1)
        self.assertEqual(failedJobs[0]['failedReason'],
                         'File /this/is/file2 has a single lumi 1, in run 1 with too many events 1000 and it would take 12000 sec to run')

        return

    def testF_HardLimitSplittingOnly(self):
        """
        _testF_HardLimitSplittingOnly_

        Checks that we can split a set of files where every file has a single
        lumi too big to fit in a runnable job
        """
        splitter = SplitterFactory()

        # Create 3 single-big-lumi files
        testFileset = Fileset(name="FilesetA")
        testFileA = self.createFile("/this/is/file1", 1000, 0, 1, "somese.cern.ch")
        testFileB = self.createFile("/this/is/file2", 1000, 1, 1, "somese.cern.ch")
        testFileC = self.createFile("/this/is/file3", 1000, 2, 1, "somese.cern.ch")
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        # Settings are to split on job boundaries, to fail sing lumis with more than 800 events
        # and to put 550 events per job
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=550,
                               job_time_limit=9600,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 3, "Three jobs must be in the jobgroup")
        for i in range(0, 3):
            self.assertTrue(jobs[i]['failedOnCreation'], "It should have been marked as failed")

            runNums = list(jobs[i]['mask']['runAndLumis'])
            self.assertEqual(len(runNums), 1)

            lumiNums = next(iter(viewvalues(jobs[i]['mask']['runAndLumis'])))
            self.assertEqual(len(lumiNums), 1)

            finalLumi = []
            for pair in lumiNums:
                finalLumi.extend(list(range(pair[0], pair[1] + 1)))
            self.assertEqual(len(finalLumi), 1)

            self.assertEqual(jobs[i]['failedReason'],
                             "File /this/is/file%d has a single lumi %s, in run %s with too many events 1000 and it would take 12000 sec to run" % (
                             i + 1, finalLumi[0], runNums[0]))

        return

    def testG_LumiMask(self):
        """
        _testG_LumiMask_

        Test that we can use a lumi-mask to filter good runs/lumis.
        """
        splitter = SplitterFactory()

        # Create 3 files with 100 events per lumi:
        # - file1 with 1 run  of 8 lumis
        # - file2 with 2 runs of 2 lumis each
        # - file3 with 1 run  of 5 lumis
        fileA = File(lfn="/this/is/file1", size=1000, events=800)
        fileB = File(lfn="/this/is/file2", size=1000, events=400)
        fileC = File(lfn="/this/is/file3", size=1000, events=500)

        lumiListA = []
        for lumi in range(8):
            lumiListA.append(10 + lumi)
        fileA.addRun(Run(1, *lumiListA))
        fileA.setLocation("somese.cern.ch")
        lumiListB1 = []
        lumiListB2 = []
        for lumi in range(2):
            lumiListB1.append(20 + lumi)
            lumiListB2.append(30 + lumi)
        fileB.addRun(Run(2, *lumiListB1))
        fileB.addRun(Run(3, *lumiListB2))
        fileB.setLocation("somese.cern.ch")
        lumiListC = []
        for lumi in range(5):
            lumiListC.append(40 + lumi)
        fileC.addRun(Run(4, *lumiListC))
        fileC.setLocation("somese.cern.ch")

        testFileset = Fileset(name='Fileset')
        testFileset.addFile(fileA)
        testFileset.addFile(fileB)
        testFileset.addFile(fileC)

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        # Use a lumi-mask = {1: [[10,14]], 2: [[20,21]], 4: [[40,41]]}
        jobGroups = jobFactory(halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               events_per_job=850,
                               runs=['1', '2', '4'],
                               lumis=['10,14', '20,21', '40,41'],
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 2, "Two jobs must be in the jobgroup")
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {1: [[10, 14]], 2: [[20, 21]], 4: [[40, 40]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {4: [[41, 41]]})

    def testH_LumiCorrections(self):
        """
        _LumiCorrections_

        Test the splitting algorithm can handle lumis which
        cross multiple files.
        """
        splitter = SplitterFactory()
        testSubscription = self.createSubscription(nFiles=2, lumisPerFile=2,
                                                   twoSites=False, nEventsPerFile=150)
        files = testSubscription.getFileset().getFiles()
        self.assertEqual(len(files), 2)
        # at the moment we have two files with two lumis each:
        #  file0 has run0 and lumis 0,1. 150 events
        #  file1 has run1 and lumis 2,3. 150 evens
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)

        jobGroups = jobFactory(events_per_job=50,
                               halt_job_on_file_boundaries=False,
                               splitOnRun=False,
                               performance=self.performanceParams,
                               applyLumiCorrection=False
                               )

        # The splitting algorithm will assume 75 events per lumi.
        # We will have one job per lumi
        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 4)

        testSubscription = self.createSubscription(nFiles=2, lumisPerFile=2,
                                                   twoSites=False, nEventsPerFile=150)
        files = testSubscription.getFileset().getFiles()
        # Now modifyng and adding duplicated lumis.
        for runObj in files[0]['runs']:
            if runObj.run != 0:
                continue
            runObj.appendLumi(42)
        for runObj in files[1]['runs']:
            if runObj.run != 1:
                continue
            runObj.run = 0
            runObj.appendLumi(42)
        files[1]['locations'] = set(['blenheim'])
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        jobGroups = jobFactory(events_per_job=50,
                               halt_job_on_file_boundaries=True,
                               performance=self.performanceParams,
                               applyLumiCorrection=True)

        # Now we will have:
        #   file0: Run0 and lumis [0, 1, 42]
        #   file1: Run0 and lumis [2, 3, 42]
        # Splitting algorithm is assuming 50 events per lumi
        # Three jobs (one per lumu) for the first file
        # Two jobs for the second file (42 is duplicated)
        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 5)
        self.assertEqual(len(jobs[0]['input_files']), 1)
        self.assertEqual(len(jobs[1]['input_files']), 1)
        self.assertEqual(len(jobs[2]['input_files']), 2)
        self.assertEqual(len(jobs[3]['input_files']), 1)
        self.assertEqual(len(jobs[4]['input_files']), 1)
        self.assertEqual(jobs[0]['mask'].getRunAndLumis(), {0: [[0, 0]]})
        self.assertEqual(jobs[1]['mask'].getRunAndLumis(), {0: [[1, 1]]})
        self.assertEqual(jobs[2]['mask'].getRunAndLumis(), {0: [[42, 42]]})
        self.assertEqual(jobs[3]['mask'].getRunAndLumis(), {0: [[2, 2]]})
        self.assertEqual(jobs[4]['mask'].getRunAndLumis(), {0: [[3, 3]]})

        # Check that if the last two jobs have the same duplicated lumi you do not get an error
        testSubscription = self.createSubscription(nFiles=2, lumisPerFile=2,
                                                   twoSites=False, nEventsPerFile=150)
        files = testSubscription.getFileset().getFiles()
        # Now modifying and adding the same duplicated lumis in the Nth and Nth-1 jobs
        for runObj in files[0]['runs']:
            if runObj.run != 0:
                continue
            runObj.appendLumi(42)
        for runObj in files[1]['runs']:
            runObj.run = 0
            runObj.lumis = [42]
        files[1]['locations'] = set(['blenheim'])
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        jobGroups = jobFactory(events_per_job=50,
                               halt_job_on_file_boundaries=True,
                               performance=self.performanceParams,
                               applyLumiCorrection=True)

        self.assertEqual(len(jobGroups), 1)
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 3)

    def testI_DisableHardLimitSplitting(self):
        """
        _testI_DisableHardLimitSplitting_

        Test that we can bypass the job time limit when allowCreationFailure is
        set to False. The algorithm shall take single lumi files with time per
        lumi greater than the job time limit but not mark them for failure
        """
        splitter = SplitterFactory()

        # Create 3 files, the one in the middle is a "bad" file
        testFileset = Fileset(name="FilesetA")
        testFileA = self.createFile("/this/is/file1", 1000, 0, 5, "blenheim")
        testFileB = self.createFile("/this/is/file2", 1000, 1, 1, "blenheim")
        testFileC = self.createFile("/this/is/file3", 1000, 2, 2, "blenheim")
        testFileset.addFile(testFileA)
        testFileset.addFile(testFileB)
        testFileset.addFile(testFileC)

        testSubscription = Subscription(fileset=testFileset,
                                        workflow=self.testWorkflow,
                                        split_algo="EventAwareLumiBased",
                                        type="Processing")
        jobFactory = splitter(package="WMCore.DataStructs",
                              subscription=testSubscription)
        # Settings are to split on job boundaries, to fail sing lumis with more than 800 events
        # and to put 550 events per job
        jobGroups = jobFactory(halt_job_on_file_boundaries=True,
                               splitOnRun=True,
                               events_per_job=550,
                               job_time_limit=9600,
                               allowCreationFailure=False,
                               performance=self.performanceParams)

        self.assertEqual(len(jobGroups), 1, "There should be only one job group")
        jobs = jobGroups[0].jobs
        self.assertEqual(len(jobs), 6, "Six jobs must be in the jobgroup")
        failedJobs = [job for job in jobs if job.get('failedOnCreation', False)]
        self.assertEqual(len(failedJobs), 0, "There should be no failed jobs")

        return

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
    CouchWorkQueueElement unit tests
"""
import unittest

import time

from Utils.PythonVersion import PY3
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit
from WMCore.WorkQueue.WorkQueueBackend import WorkQueueBackend, sortAvailableElements
from WMCore.WorkQueue.DataStructs.CouchWorkQueueElement import CouchWorkQueueElement
from WMCore.WorkQueue.DataStructs.WorkQueueElement import WorkQueueElement

from WMCore.WMSpec.StdSpecs.ReReco import ReRecoWorkloadFactory
from WMQuality.Emulators.WMSpecGenerator.WMSpecGenerator import createConfig

rerecoArgs = ReRecoWorkloadFactory.getTestArguments()


def rerecoWorkload(workloadName, arguments):
    factory = ReRecoWorkloadFactory()
    wmspec = factory.factoryWorkloadConstruction(workloadName, arguments)
    return wmspec


class WorkQueueBackendTest(unittest.TestCase):
    def setUp(self):
        self.testInit = TestInit('CouchWorkQueueTest')
        self.testInit.setLogging()
        self.testInit.setupCouch('wq_backend_test_inbox', 'WorkQueue')
        self.testInit.setupCouch('wq_backend_test', 'WorkQueue')
        self.testInit.setupCouch('wq_backend_test_parent', 'WorkQueue')
        self.couch_db = self.testInit.couch.couchServer.connectDatabase('wq_backend_test')
        self.backend = WorkQueueBackend(db_url=self.testInit.couchUrl,
                                        db_name='wq_backend_test',
                                        inbox_name='wq_backend_test_inbox',
                                        parentQueue='%s/%s' % (self.testInit.couchUrl, 'wq_backend_test_parent'))
        rerecoArgs["ConfigCacheID"] = createConfig(rerecoArgs["CouchDBName"])
        self.processingSpec = rerecoWorkload('testProcessing', rerecoArgs)
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def tearDown(self):
        """
        _tearDown_

        """
        self.testInit.tearDownCouch()

    def testPriority(self):
        """Element priority and ordering handled correctly"""
        element = WorkQueueElement(RequestName='backend_test',
                                   WMSpec=self.processingSpec,
                                   Status='Available',
                                   SiteWhitelist=["place"],
                                   Jobs=10, Priority=1)
        highprielement = WorkQueueElement(RequestName='backend_test_high',
                                          WMSpec=self.processingSpec,
                                          Status='Available', Jobs=10,
                                          SiteWhitelist=["place"],
                                          Priority=100)
        element2 = WorkQueueElement(RequestName='backend_test_2',
                                    WMSpec=self.processingSpec,
                                    Status='Available',
                                    SiteWhitelist=["place"],
                                    Jobs=10, Priority=1)
        element3 = WorkQueueElement(RequestName='backend_test_3',
                                    WMSpec=self.processingSpec,
                                    Status='Available',
                                    SiteWhitelist=["place"],
                                    Jobs=10, Priority=1)
        lowprielement = WorkQueueElement(RequestName='backend_test_low',
                                         WMSpec=self.processingSpec,
                                         Status='Available',
                                         SiteWhitelist=["place"],
                                         Jobs=10, Priority=0.1)
        self.backend.insertElements([element])
        self.backend.availableWork({'place': 1000}, {})
        # timestamp in elements have second coarseness, 2nd element must
        # have a higher timestamp to force it after the 1st
        time.sleep(1)
        self.backend.insertElements([lowprielement, element2, highprielement])
        self.backend.availableWork({'place': 1000}, {})
        time.sleep(1)
        self.backend.insertElements([element3])
        work = self.backend.availableWork({'place': 1000}, {})
        # order should be high to low, with the standard elements in the order
        # they were queueud
        self.assertEqual([x['RequestName'] for x in work[0]],
                         ['backend_test_high', 'backend_test', 'backend_test_2',
                          'backend_test_3', 'backend_test_low'])

    def testDuplicateInsertion(self):
        """Try to insert elements multiple times"""
        element1 = CouchWorkQueueElement(self.couch_db,
                                         elementParams={'RequestName': 'backend_test',
                                                        'WMSpec': self.processingSpec,
                                                        'Status': 'Available',
                                                        'Jobs': 10,
                                                        'Inputs': {
                                                            self.processingSpec.listInputDatasets()[0] + '#1': []}})
        element2 = CouchWorkQueueElement(self.couch_db,
                                         elementParams={'RequestName': 'backend_test',
                                                        'WMSpec': self.processingSpec,
                                                        'Status': 'Available',
                                                        'Jobs': 20,
                                                        'Inputs': {
                                                            self.processingSpec.listInputDatasets()[0] + '#2': []}})
        self.backend.insertElements([element1, element2])
        self.backend.insertElements([element1, element2])
        # check no duplicates and no conflicts
        self.assertEqual(len(self.backend.db.allDocs()['rows']), 4)  # design doc + workflow + 2 elements
        self.assertEqual(self.backend.db.loadView('WorkQueue', 'conflicts')['total_rows'], 0)

    def testSortAvailableElements(self):
        """Test logic used elemets sorting, in the fuction `sortAvailableElements`"""
        expected = [{'CreationTime': 2.4, 'Jobs': 2, 'Priority': 200},
                    {'CreationTime': 4.3, 'Jobs': 3, 'Priority': 200},
                    {'CreationTime': 1.5, 'Jobs': 5, 'Priority': 100},
                    {'CreationTime': 3.1, 'Jobs': 1, 'Priority': 100},
                    {'CreationTime': 5.2, 'Jobs': 4, 'Priority': 100}]

        elemList = [{'CreationTime': 3.1, 'Jobs': 1, 'Priority': 100},
                    {'CreationTime': 2.4, 'Jobs': 2, 'Priority': 200},
                    {'CreationTime': 4.3, 'Jobs': 3, 'Priority': 200},
                    {'CreationTime': 5.2, 'Jobs': 4, 'Priority': 100},
                    {'CreationTime': 1.5, 'Jobs': 5, 'Priority': 100}]
        sortAvailableElements(elemList)
        self.assertEqual(len(elemList), 5)
        self.assertItemsEqual(elemList, expected)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_Harvest_t_

Harvest job splitting test

"""

from builtins import str, range

import unittest
import threading
import logging
import time
import os

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.DataStructs.LumiList import LumiList
from WMCore.DataStructs.Run import Run
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.DAOFactory import DAOFactory
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.ResourceControl.ResourceControl import ResourceControl
from WMQuality.TestInit import TestInit
from WMQuality.Emulators import EmulatorSetup


def createCommonFileset():
    """
    Create a simple fileset with 2 files at the same location
    """
    multipleFilesFileset = Fileset(name="TestFileset")

    newFile = File("/some/file/test1", size=1000, events=100)
    newFile.addRun(Run(1, *[1, 3, 4, 5, 6, 7]))
    newFile.addRun(Run(2, *[1, 2, 4, 5, 6, 7]))
    newFile.setLocation('T2_CH_CERN')
    multipleFilesFileset.addFile(newFile)

    newFile = File("/some/file/test2", size=2000, events=200)
    newFile.addRun(Run(3, *[2, 8]))
    newFile.addRun(Run(4, *[3, 8]))
    newFile.setLocation('T2_CH_CERN')
    multipleFilesFileset.addFile(newFile)

    newFile = File("/some/file/test3", size=3000, events=300)
    newFile.addRun(Run(5, *[10, 11, 12]))
    newFile.addRun(Run(6, *[10, 11, 12]))
    newFile.setLocation('T2_CH_CERN')
    multipleFilesFileset.addFile(newFile)

    newFile = File("/some/file/test4", size=4000, events=400)
    newFile.addRun(Run(2, *[3, 8, 9]))
    newFile.addRun(Run(3, *[3, 4, 5, 6]))
    newFile.setLocation('T2_CH_CERN')
    multipleFilesFileset.addFile(newFile)

    multipleFilesFileset.create()
    return multipleFilesFileset


class HarvestTest(unittest.TestCase):
    """
    _HarvestTest_

    Test for EndOfRun job splitter
    """

    def setUp(self):
        """
        _setUp_

        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()

        self.testInit.setSchema(customModules=["WMCore.WMBS"])

        self.splitterFactory = SplitterFactory(package="WMCore.JobSplitting")

        myThread = threading.currentThread()
        self.myThread = myThread
        daoFactory = DAOFactory(package="WMCore.WMBS",
                                logger=logging,
                                dbinterface=myThread.dbi)
        self.WMBSFactory = daoFactory

        config = self.getConfig()
        self.changer = ChangeState(config)

        myResourceControl = ResourceControl()
        myResourceControl.insertSite("T1_US_FNAL", 10, 20, "T1_US_FNAL_Disk", "T1_US_FNAL")
        myResourceControl.insertSite("T1_US_FNAL", 10, 20, "T3_US_FNALLPC", "T1_US_FNAL")
        myResourceControl.insertSite("T2_CH_CERN", 10, 20, "T2_CH_CERN", "T2_CH_CERN")

        self.fileset1 = Fileset(name="TestFileset1")
        for fileNum in range(11):
            newFile = File("/some/file/name%d" % fileNum, size=1000, events=100)
            newFile.addRun(Run(1, *[1]))
            newFile.setLocation('T1_US_FNAL_Disk')
            self.fileset1.addFile(newFile)

        self.fileset1.create()

        workflow1 = Workflow(spec="spec.xml", owner="hufnagel", name="TestWorkflow1", task="Test")
        workflow1.create()

        self.subscription1 = Subscription(fileset=self.fileset1,
                                          workflow=workflow1,
                                          split_algo="Harvest",
                                          type="Harvesting")

        self.subscription1.create()
        self.configFile = EmulatorSetup.setupWMAgentConfig()

        return

    def tearDown(self):
        """
        _tearDown_

        """
        self.testInit.clearDatabase()
        EmulatorSetup.deleteConfig(self.configFile)

        return

    def getConfig(self):
        """
        _getConfig_

        """
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        config.section_("CoreDatabase")
        config.CoreDatabase.connectUrl = os.getenv("DATABASE")
        config.CoreDatabase.socket = os.getenv("DBSOCK")

        # JobStateMachine
        config.component_('JobStateMachine')
        config.JobStateMachine.couchurl = os.getenv('COUCHURL', None)
        config.JobStateMachine.couchDBName = 'wmagent_jobdump'

        return config

    def finishJobs(self, jobGroups, subscription=None):
        """
        _finishJobs_

        """
        if not subscription:
            subscription = self.subscription1
        for f in subscription.acquiredFiles():
            subscription.completeFiles(f)

        for jobGroup in jobGroups:
            self.changer.propagate(jobGroup.jobs, 'executing', 'created')
            self.changer.propagate(jobGroup.jobs, 'complete', 'executing')
            self.changer.propagate(jobGroup.jobs, 'success', 'complete')
            self.changer.propagate(jobGroup.jobs, 'cleanout', 'success')

        return

    def testHarvestEndOfRunTrigger(self):
        """
        _testDQMHarvestEndOfRunTrigger_

        Make sure that the basic splitting algo works, which is only, ExpressMerge is ALL done, fire a job against that fileset

        """
        self.assertEqual(self.fileset1.open, True, "Fileset is closed. Shouldn't")

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)

        jobGroups = jobFactory()

        self.assertEqual(len(jobGroups), 0,
                         "We got 1 or more jobGroups with an open fileset and no periodic configuration")

        self.fileset1.markOpen(False)
        self.assertEqual(self.fileset1.open, False, "Fileset is opened, why?")

        # We should also check if there are aqcuired files, if there are, there are jobs,
        # we don't want to fire another jobs while previous are running (output is integrating whatever  input)
        # TODO : The above one we can do when all is done. Not priority

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)
        jobGroups = jobFactory()

        self.assertEqual(len(jobGroups), 1,
                         "Harvest jobsplitter didn't create a single jobGroup after the fileset was closed")

        return

    def testPeriodicTrigger(self):
        """
        _testPeriodicTrigger_

        """
        self.assertEqual(self.fileset1.open, True, "Fileset is not open, not testing periodic here")
        # Test timeout (5s for this first test)
        # there should be no acquired files, if there are, shouldn't be a job
        # self.subscription1.acquireFiles(self.subscription1.availableFiles().pop())

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)
        jobGroups = jobFactory(periodic_harvest_interval=3)

        self.assertEqual(len(jobGroups), 1, "Didn't created the first periodic job when there were acquired files")

        # For the whole thing to work, faking the first job finishing, and putting the files as complete
        self.finishJobs(jobGroups)

        # Adding more of files, so we have new stuff to process
        for fileNum in range(12, 24):
            newFile = File("/some/file/name%d" % fileNum, size=1000, events=100)
            newFile.addRun(Run(1, *[1]))
            newFile.setLocation('T1_US_FNAL_Disk')
            self.fileset1.addFile(newFile)
        self.fileset1.commit()

        # Testing that it doesn't create a job unless the delay is past
        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)
        jobGroups = jobFactory(periodic_harvest_interval=2)

        self.assertEqual(len(jobGroups), 0,
                         "Created one or more job, when there were non-acquired file and the period is not passed by")

        time.sleep(2)

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)
        jobGroups = jobFactory(periodic_harvest_interval=2)

        self.assertEqual(len(jobGroups), 1,
                         "Didn't created one or more job, and there weren't and the period is passed by")

        # Finishing out previous jobs
        self.finishJobs(jobGroups)

        # Adding more of files, so we have new stuff to process
        for fileNum in range(26, 36):
            newFile = File("/some/file/name%d" % fileNum, size=1000, events=100)
            newFile.addRun(Run(1, *[1]))
            newFile.setLocation('T1_US_FNAL_Disk')
            self.fileset1.addFile(newFile)
        self.fileset1.commit()

        # Trying to create another job just afterwards, it shouldn't, because it should respect the configured delay
        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)
        jobGroups = jobFactory(periodic_harvest_interval=2)

        self.assertEqual(len(jobGroups), 0, "Created one or more job, there are new files, but the delay is not past")

        time.sleep(2)

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)
        jobGroups = jobFactory(periodic_harvest_interval=2)

        self.assertEqual(len(jobGroups), 1, "Didn't created one or more job, there are new files and the delay is past")

        # Last check is whether the job gets all the files or not

        numFilesJob = jobGroups[0].jobs[0].getFiles()
        numFilesFileset = self.fileset1.getFiles()
        self.assertEqual(numFilesJob, numFilesFileset, "Job didn't got all the files")

        # Finishing out previous jobs
        self.finishJobs(jobGroups)

        # Adding files for the first location
        for fileNum in range(38, 48):
            newFile = File("/some/file/name%d" % fileNum, size=1000, events=100)
            newFile.addRun(Run(1, *[1]))
            newFile.setLocation('T1_US_FNAL_Disk')
            self.fileset1.addFile(newFile)
        self.fileset1.commit()
        # Then another location
        for fileNum in range(50, 56):
            newFile = File("/some/file/name%d" % fileNum, size=1000, events=100)
            newFile.addRun(Run(1, *[1]))
            newFile.setLocation('T2_CH_CERN')
            self.fileset1.addFile(newFile)
        self.fileset1.commit()

        # We should have jobs in both locations
        time.sleep(2)

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)
        jobGroups = jobFactory(periodic_harvest_interval=2)

        self.assertEqual(len(jobGroups[0].getJobs()), 2, "We didn't get 2 jobs for 2 locations")

        firstJobLocation = jobGroups[0].getJobs()[0].getFileLocations()[0]
        secondJobLocation = jobGroups[0].getJobs()[1].getFileLocations()[0]

        self.assertEqual(firstJobLocation, 'T2_CH_CERN')
        self.assertEqual(secondJobLocation, 'T1_US_FNAL')

        self.finishJobs(jobGroups)

        for fileNum in range(60, 65):
            newFile = File("/some/file/name%d" % fileNum, size=1000, events=100)
            newFile.addRun(Run(2, *[2]))
            newFile.setLocation('T2_CH_CERN')
            self.fileset1.addFile(newFile)
        self.fileset1.commit()

        for fileNum in range(70, 75):
            newFile = File("/some/file/name%d" % fileNum, size=1000, events=100)
            newFile.addRun(Run(3, *[3]))
            newFile.setLocation('T2_CH_CERN')
            self.fileset1.addFile(newFile)
        self.fileset1.commit()

        time.sleep(2)

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=self.subscription1)
        jobGroups = jobFactory(periodic_harvest_interval=2)

        # This is one of the most "complicated" tests so worth to comment, 4 jobs should be created
        # 1 - all previous files from SomeSE and run = 1 (a lot, like ~45)
        # 2 - Few files from SomeSE3, Run = 1
        # 3 - Few files from SomeSE3, Run = 2
        # 4 - Few files from SomeSE3, Run = 3
        self.assertEqual(len(jobGroups[0].getJobs()), 4, "We didn't get 4 jobs for adding 2 different runs to SomeSE3")

        return

    def testMultipleRunHarvesting(self):
        """
        _testMultipleRunHarvesting_

        Add some files with multiple runs in each, make sure the jobs
        are created by location and run. Verify each job mask afterwards.
        Note that in this test run are splitted between sites,
        in real life that MUST NOT happen we still don't support that.
        """
        multipleFilesFileset = Fileset(name="TestFileset")

        newFile = File("/some/file/test1", size=1000, events=100)
        newFile.addRun(Run(1, *[1, 3, 4, 5, 6, 7]))
        newFile.addRun(Run(2, *[1, 2, 4, 5, 6, 7]))
        newFile.setLocation('T1_US_FNAL_Disk')
        multipleFilesFileset.addFile(newFile)
        newFile = File("/some/file/test2", size=1000, events=100)
        newFile.addRun(Run(1, *[2, 8]))
        newFile.addRun(Run(2, *[3, 8]))
        newFile.setLocation('T2_CH_CERN')
        multipleFilesFileset.addFile(newFile)
        multipleFilesFileset.create()

        harvestingWorkflow = Workflow(spec="spec.xml",
                                      owner="hufnagel",
                                      name="TestWorkflow",
                                      task="Test")
        harvestingWorkflow.create()

        harvestSub = Subscription(fileset=multipleFilesFileset,
                                  workflow=harvestingWorkflow,
                                  split_algo="Harvest",
                                  type="Harvesting")
        harvestSub.create()

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=harvestSub)
        jobGroups = jobFactory(periodic_harvest_interval=2)
        self.assertEqual(len(jobGroups), 1, "A single job group was not created")
        self.assertEqual(len(jobGroups[0].getJobs()), 4,
                         "Four jobs were not created")

        for job in jobGroups[0].getJobs():
            runs = job['mask'].getRunAndLumis()
            self.assertEqual(len(runs), 1, "Job has more than one run configured")
            ll = LumiList(compactList={1: [[1, 1], [3, 7], [2, 2], [8, 8]],
                                       2: [[1, 2], [4, 7], [3, 3], [8, 8]]})
            run = next(iter(runs))
            for lumiPair in runs[run]:
                for lumi in range(lumiPair[0], lumiPair[1] + 1):
                    self.assertTrue((str(run), lumi) in ll, "All of %s not in %s" % (lumiPair, ll))

        self.finishJobs(jobGroups, harvestSub)

        newFile = File("/some/file/test3", size=1000, events=100)
        newFile.addRun(Run(1, *list(range(9, 15))))
        newFile.setLocation('T2_CH_CERN')
        multipleFilesFileset.addFile(newFile)
        multipleFilesFileset.commit()

        time.sleep(2)

        jobGroups = jobFactory(periodic_harvest_interval=2)
        self.assertEqual(len(jobGroups), 1, "A single job group was not created")
        self.assertEqual(len(jobGroups[0].getJobs()), 4, "Four jobs were not created")

        for job in jobGroups[0].getJobs():
            runs = job['mask'].getRunAndLumis()
            self.assertEqual(len(runs), 1, "Job has more than one run configured")
            ll = LumiList(compactList={1: [[1, 1], [3, 7], [2, 2], [8, 8], [9, 14]],
                                       2: [[1, 2], [4, 7], [3, 3], [8, 8]]})
            run = next(iter(runs))
            for lumiPair in runs[run]:
                for lumi in range(lumiPair[0], lumiPair[1] + 1):
                    self.assertTrue((run, lumi) in ll, "All of %s not in %s" % (lumiPair, ll))

        harvestingWorkflowSib = Workflow(spec="spec.xml",
                                         owner="hufnagel",
                                         name="TestWorkflowSib",
                                         task="TestSib")
        harvestingWorkflowSib.create()

        harvestSubSib = Subscription(fileset=multipleFilesFileset,
                                     workflow=harvestingWorkflowSib,
                                     split_algo="Harvest",
                                     type="Harvesting")
        harvestSubSib.create()

        jobFactorySib = self.splitterFactory(package="WMCore.WMBS", subscription=harvestSubSib)

        multipleFilesFileset.markOpen(False)

        jobGroups = jobFactorySib(periodic_harvest_sibling=True)
        self.assertEqual(len(jobGroups), 0, "A single job group was created")

        self.finishJobs(jobGroups, harvestSub)

        jobGroups = jobFactorySib(periodic_harvest_sibling=True)
        self.assertEqual(len(jobGroups), 1, "A single job group was not created")
        self.assertEqual(len(jobGroups[0].getJobs()), 4, "Four jobs were not created")

        for job in jobGroups[0].getJobs():
            runs = job['mask'].getRunAndLumis()
            self.assertEqual(len(runs), 1, "Job has more than one run configured")
            ll = LumiList(compactList={1: [[1, 1], [3, 7], [2, 2], [8, 8], [9, 14]],
                                       2: [[1, 2], [4, 7], [3, 3], [8, 8]]})
            run = next(iter(runs))
            for lumiPair in runs[run]:
                for lumi in range(lumiPair[0], lumiPair[1] + 1):
                    self.assertTrue((run, lumi) in ll, "All of %s not in %s" % (lumiPair, ll))

    def testMultiRunHarvesting(self):
        """
        _testMultiRunHarvesting_

        Provided a fileset with a couple of files and different runs, create a
        single job for all the runs at a specific location, which also adds a
        baggage to the job (True) which is later on looked up by SetupCMSSWPSet.
        """
        multipleFilesFileset = createCommonFileset()
        self.assertEqual(multipleFilesFileset.open, True)

        harvestingWorkflow = Workflow(spec="spec.xml",
                                      owner="amaltaro",
                                      name="TestWorkflow",
                                      task="Test")
        harvestingWorkflow.create()

        harvestSub = Subscription(fileset=multipleFilesFileset,
                                  workflow=harvestingWorkflow,
                                  split_algo="Harvest",
                                  type="Harvesting")
        harvestSub.create()

        multipleFilesFileset.markOpen(False)
        self.assertEqual(multipleFilesFileset.open, False, "Fileset should now be closed")

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=harvestSub)
        jobGroups = jobFactory(dqmHarvestUnit="multiRun")
        self.assertEqual(len(jobGroups), 1)

        for jobGroup in jobGroups:
            self.assertEqual(len(jobGroup.jobs), 1)
            for job in jobGroup.jobs:
                baggage = job.getBaggage()
                self.assertTrue(getattr(baggage, "multiRun", False), "It's supposed to be a multiRun job")
                self.assertEqual(getattr(baggage, "runLimits", ""), "-1-6")

    def testByRunHarvesting(self):
        """
        _testByRunHarvesting_
        Provided a fileset with a couple of files and 4 different runs, create
        one single job per run and location.
        The multiRun baggage should be false in this case.
        """
        multipleFilesFileset = createCommonFileset()
        self.assertEqual(multipleFilesFileset.open, True, "Fileset should be open!")

        harvestingWorkflow = Workflow(spec="spec.xml",
                                      owner="amaltaro",
                                      name="TestWorkflow",
                                      task="Test")
        harvestingWorkflow.create()

        harvestSub = Subscription(fileset=multipleFilesFileset,
                                  workflow=harvestingWorkflow,
                                  split_algo="Harvest",
                                  type="Harvesting")
        harvestSub.create()

        multipleFilesFileset.markOpen(False)
        self.assertEqual(multipleFilesFileset.open, False, "Fileset should now be closed")

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=harvestSub)
        jobGroups = jobFactory()
        self.assertEqual(len(jobGroups), 1, "Should have created 1 job group")

        for jobGroup in jobGroups:
            self.assertEqual(len(jobGroup.jobs), 6, "Should have created 6 jobs")
            for job in jobGroup.jobs:
                baggage = job.getBaggage()
                self.assertFalse(getattr(baggage, "multiRun", False), "It's supposed to be a byRun job")

    def testByRunAndRunWhitelist(self):
        """
        _testByRunAndRunWhitelist_

        Create harvesting jobs by run for the runs provided in the RunWhitelist
        """
        multipleFilesFileset = createCommonFileset()
        self.assertEqual(multipleFilesFileset.open, True)

        harvestingWorkflow = Workflow(spec="spec.xml", owner="amaltaro",
                                      name="TestWorkflow", task="Test")
        harvestingWorkflow.create()

        harvestSub = Subscription(fileset=multipleFilesFileset, workflow=harvestingWorkflow,
                                  split_algo="Harvest", type="Harvesting")
        harvestSub.create()

        multipleFilesFileset.markOpen(False)
        self.assertEqual(multipleFilesFileset.open, False, "Fileset should now be closed")

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=harvestSub)
        jobGroups = jobFactory(runWhitelist=[1, 3])
        self.assertEqual(len(jobGroups), 1, "One jobgroup per location")

        for jobGroup in jobGroups:
            self.assertEqual(len(jobGroup.jobs), 2)

    def testByRunAndRunBlacklist(self):
        """
        _testByRunAndRunWhitelist_

        Create harvesting jobs by run for the runs provided in the RunWhitelist
        """
        multipleFilesFileset = createCommonFileset()
        self.assertEqual(multipleFilesFileset.open, True)

        harvestingWorkflow = Workflow(spec="spec.xml", owner="amaltaro",
                                      name="TestWorkflow", task="Test")
        harvestingWorkflow.create()

        harvestSub = Subscription(fileset=multipleFilesFileset, workflow=harvestingWorkflow,
                                  split_algo="Harvest", type="Harvesting")
        harvestSub.create()

        multipleFilesFileset.markOpen(False)
        self.assertEqual(multipleFilesFileset.open, False, "Fileset should now be closed")

        jobFactory = self.splitterFactory(package="WMCore.WMBS", subscription=harvestSub)
        jobGroups = jobFactory(runWhitelist=[1, 2, 3, 4, 5], runBlacklist=[1, 3])
        self.assertEqual(len(jobGroups), 1, "One jobgroup per location")

        for jobGroup in jobGroups:
            self.assertEqual(len(jobGroup.jobs), 3)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_SetBlockFiles_

MySQL implementation of DBSBufferFiles.SetBlock
"""

from WMCore.Database.DBFormatter import DBFormatter

class SetBlockFiles(DBFormatter):
    """
    Update files with block info
    """
    sql = """UPDATE dbsbuffer_file SET block_id =
               (SELECT id FROM dbsbuffer_block WHERE blockname = :block)
             WHERE dbsbuffer_file.lfn = :filelfn"""

    def execute(self, binds, conn = None, transaction = None):
        """
        Insert block name into files

        Requires you to pass in binds correctly
        """

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



"""
File       : MSPileupMonitoring.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: MSPileupMonitoring provides bridge between MSPileup
             service and CMS Monitoring infrastructure
"""

# system modules
import time

# WMCore modules
from WMCore.MicroService.Tools.Common import getMSLogger

# CMSMonitoring modules
from CMSMonitoring.StompAMQ7 import StompAMQ7 as StompAMQ


def flatDocuments(doc):
    """
    Helper function to flat out MSPileup document

    :param doc: input MSPileup document
    :return: generator of MSPileup documents flatten from original one
    """
    docs = flatKey(doc, 'campaigns')
    docs = (f for d in docs for f in flatKey(d, 'currentRSEs'))
    docs = (f for d in docs for f in flatKey(d, 'expectedRSEs'))
    for doc in docs:
        yield doc


def flatKey(doc, key):
    """
    Helper function to flat out values of given key in a document

    :param doc: input MSPileup document
    :param key: document key to use
    :return: generator of MSPileup documents flatten from original one and given key
    """
    for item in doc[key]:
        ndoc = dict(doc)
        # convert plural to singular key name, e.g. campaigns -> campaign
        nkey = key[:-1]
        ndoc[nkey] = item
        del ndoc[key]
        yield ndoc


class MSPileupMonitoring():
    """
    MSPileupMonitoring represents MSPileup monitoring class
    """

    def __init__(self, msConfig=None):
        """
        Constructor for MSPileupMonitoring
        """
        self.userAMQ = msConfig.get('user_amq', None)
        self.passAMQ = msConfig.get('pass_amq', None)
        self.topicAMQ = msConfig.get('topic_amq', None)
        self.docTypeAMQ = msConfig.get('doc_type_amq', 'cms-ms-pileup')
        self.hostPortAMQ = msConfig.get('host_port_amq', None)
        self.producer = msConfig.get('producer', 'cms-ms-pileup')
        self.logger = msConfig.get('logger', getMSLogger(False))

    def uploadToAMQ(self, docs, producer=None):
        """
        _uploadToAMQ_

        Sends data to AMQ, which ends up in elastic search.
        :param docs: list of documents/dicts to be posted
        :param producer: service name that's providing this info
        :return: {} or {"success": ndocs, "failures": nfailures}
        """
        if not docs:
            self.logger.info("There are no documents to send to AMQ")
            return {}
        if not self.userAMQ or not self.passAMQ:
            self.logger.info("MSPileupMonitoring has no AMQ credentials, will skip the upload to MONIT")
            return {}

        producer = producer or self.producer
        ts = int(time.time())
        notifications = []

        self.logger.debug("Sending %d to AMQ", len(docs))
        try:
            stompSvc = StompAMQ(username=self.userAMQ,
                                password=self.passAMQ,
                                producer=producer,
                                topic=self.topicAMQ,
                                validation_schema=None,
                                host_and_ports=self.hostPortAMQ,
                                logger=self.logger)

            for doc in docs:
                singleNotif, _, _ = stompSvc.make_notification(payload=doc, doc_type=self.docTypeAMQ,
                                                               ts=ts, data_subfield="payload")
                notifications.append(singleNotif)

            failures = stompSvc.send(notifications)
            msg = "%i out of %i documents successfully sent to AMQ" % (len(notifications) - len(failures),
                                                                       len(notifications))
            self.logger.info(msg)
            return {"success": len(notifications) - len(failures), "failures": len(failures)}
        except Exception as ex:
            self.logger.exception("Failed to send data to StompAMQ. Error %s", str(ex))
        return {}
